"""
Integration tests for batch opening of channels.
"""
from typing import Dict
from unittest import TestCase, mock

import networkx as nx

from lndmanage.lib.network import Network
from lndmanage.lib.rating import ChannelRater
from lndmanage.lib.pathfinding import dijkstra

from graph_definitions.routing_graph import nodes as test_graph


def new_test_graph(graph: Dict):
    # we need to init the node interface with a public key
    class MockNode:
        pub_key = 'A'

    # we disable cached graph reading
    with mock.patch.object(Network, 'load_graph', return_value=None):
        network = Network(MockNode())
        network.graph = nx.MultiGraph()
        network.edges = {}

    # add nodes
    for node, node_definition in graph.items():
        network.graph.add_node(
            node,
            alias=node,
            last_update=None,
            address=None,
            color=None)

    # add channels
    for node, node_definition in graph.items():
        for channel, channel_definition in node_definition['channels'].items():
            # create a dictionary for channel_id lookups
            to_node = channel_definition['to']
            network.edges[channel] = {
                'node1_pub': node,
                'node2_pub': to_node,
                'capacity': channel_definition['capacity'],
                'last_update': None,
                'channel_id': channel,
                'chan_point': channel,
                'policies': {
                    node > to_node: channel_definition['policies'][node > to_node],
                    to_node > node: channel_definition['policies'][to_node > node],
                }
            }

            # add vertices to network graph for edge-based lookups
            network.graph.add_edge(
                node,
                to_node,
                channel_id=channel,
                last_update=None,
                capacity=channel_definition['capacity'],
                fees={
                    node > to_node: channel_definition['policies'][node > to_node],
                    to_node > node: channel_definition['policies'][to_node > node],
                })

    return network


class TestGraph(TestCase):
    def test_network(self):
        n = new_test_graph(test_graph)
        self.assertEqual(5, n.graph.number_of_nodes())
        self.assertEqual(7, n.graph.number_of_edges())

    def test_shortest_path(self):
        network = new_test_graph(test_graph)
        cr = ChannelRater(network)
        amt_msat = 1_000_000
        weight_function = lambda v, u, e: cr.node_to_node_weight(v, u, e, amt_msat)
        print(dijkstra(network.graph, 'A', 'E', weight=weight_function))
        # TODO: use too high capacity
        # TODO: use parallel channels with different policies

    def test_liquidity_hints(self):
        """
                3
            A  ---  B
            |    2/ |
          6 |   E   | 1
            | /5 \7 |
            D  ---  C
                4
        """
        amt_msat = 100_000 * 1_000
        network = new_test_graph(test_graph)
        cr = ChannelRater(network=network)
        weight_function = lambda v, u, e: cr.node_to_node_weight(v, u, e, amt_msat)

        path = dijkstra(network.graph, 'A', 'E', weight=weight_function)
        self.assertEqual(['A', 'D', 'E'], path)

        # We report that B cannot send to E
        network.liquidity_hints.update_cannot_send('B', 'E', 2, 1_000)
        path = dijkstra(network.graph, 'A', 'E', weight=weight_function)
        self.assertEqual(['A', 'D', 'E'], path)

        # We report that D cannot send to E
        network.liquidity_hints.update_cannot_send('D', 'E', 5, 1_000)
        path = dijkstra(network.graph, 'A', 'E', weight=weight_function)
        self.assertEqual(['A', 'D', 'C', 'E'], path)

        # We report that D can send to C
        network.liquidity_hints.update_can_send('D', 'C', 4, amt_msat + 1000)
        path = dijkstra(network.graph, 'A', 'E', weight=weight_function)
        self.assertEqual(['A', 'D', 'C', 'E'], path)



"""
Module for printing lightning channels.
"""

import math
import logging
from collections import OrderedDict
from typing import TYPE_CHECKING, Tuple, List, Dict
import time

if TYPE_CHECKING:
    from lndmanage.lib.node import LndNode

from lndmanage.lib.forwardings import (
    get_channel_properties,
    get_node_properites,
)
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# define symbols for bool to string conversion
POSITIVE_MARKER = "\u2713"
NEGATIVE_MARKER = "\u2717"
ALIAS_LENGTH = 25
ANNOTATION_LENGTH = 25

PRINT_CHANNELS_FORMAT = {
    "act": {
        "dict_key": "active",
        "description": "channel is active",
        "width": 3,
        "format": "^3",
        "align": ">",
        "convert": lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    "age": {
        "dict_key": "age",
        "description": "channel age [days]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "alias": {
        "dict_key": "alias",
        "description": "alias",
        "width": ALIAS_LENGTH,
        "format": ".<" + str(ALIAS_LENGTH),
        "align": "^",
        "convert": lambda x: alias_cutoff(x),
    },
    "annotation": {
        "dict_key": "annotation",
        "description": "channel annotation",
        "width": ANNOTATION_LENGTH,
        "format": ".<" + str(ANNOTATION_LENGTH),
        "align": "^",
        "convert": lambda x: alias_cutoff(x),
    },
    "atb": {
        "dict_key": "amount_to_balanced",
        "description": "amount to be balanced (local side) [sat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "bwd": {
        "dict_key": "bandwidth_demand",
        "description": "bandwidth demand: capacity / max(mean_in, mean_out)",
        "width": 5,
        "format": "5.2f",
        "align": ">",
    },
    "cap": {
        "dict_key": "capacity",
        "description": "channel capacity [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "cid": {
        "dict_key": "chan_id",
        "description": "channel id",
        "width": 18,
        "format": "",
        "align": "^",
    },
    "fees": {
        "dict_key": "fees_out",
        "description": "total fees [sat]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "fio": {
        "dict_key": "fees_in_out",
        "description": "total fees in and out [sat]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "fo/w": {
        "dict_key": "fees_out_per_week",
        "description": "total fees out per week [sat / week]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "fi/w": {
        "dict_key": "fees_in_per_week",
        "description": "total fees in per week [sat / week]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "nfwd/a": {
        "dict_key": "forwardings_per_channel_age",
        "description": "number of forwardings per channel age in forwarding interval [1 / days]",
        "width": 6,
        "format": "6.2f",
        "align": ">",
    },
    "flow": {
        "dict_key": "flow_direction",
        "description": "flow direction (positive is outwards)",
        "width": 5,
        "format": "5.2f",
        "align": ">",
        # 'convert': lambda x: '>'*int(nan_to_zero(x)*10/3.0) if x > 0 else
        # '<'*(-int(nan_to_zero(x)*10/3.0))
    },
    "pbf": {
        "dict_key": "peer_base_fee",
        "description": "peer base fee [msat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "pfr": {
        "dict_key": "peer_fee_rate",
        "description": "peer fee rate",
        "width": 8,
        "format": "1.6f",
        "align": ">",
        "convert": lambda x: x / 1e6,
    },
    "lbf": {
        "dict_key": "local_base_fee",
        "description": "local base fee [msat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lfr": {
        "dict_key": "local_fee_rate",
        "description": "local fee rate [sat/sat]",
        "width": 8,
        "format": "1.6f",
        "align": ">",
        "convert": lambda x: x / 1e6,
    },
    "lup": {
        "dict_key": "last_update",
        "description": "last update time [days ago]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lupp": {
        "dict_key": "last_update_peer",
        "description": "last update time by peer [days ago]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lupl": {
        "dict_key": "last_update_local",
        "description": "last update time by local [days ago]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lb": {
        "dict_key": "local_balance",
        "description": "local balance [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "nfwd": {
        "dict_key": "number_forwardings",
        "description": "number of forwardings",
        "width": 4,
        "format": "4.0f",
        "align": ">",
    },
    "priv": {
        "dict_key": "private",
        "description": "channel is private",
        "width": 5,
        "format": "^5",
        "align": ">",
        "convert": lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    "r": {
        "dict_key": "action_required",
        "description": "action is required",
        "width": 1,
        "format": "^1",
        "align": ">",
        "convert": lambda x: NEGATIVE_MARKER if x else "",
    },
    "rb": {
        "dict_key": "remote_balance",
        "description": "remote balance [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "in": {
        "dict_key": "total_forwarding_in",
        "description": "total forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "ini": {
        "dict_key": "initiator",
        "description": "true if we opened channel",
        "width": 3,
        "format": "^3",
        "align": ">",
        "convert": lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    "tot": {
        "dict_key": "total_forwarding",
        "description": "total forwarding [sat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "ub": {
        "dict_key": "unbalancedness",
        "description": "unbalancedness [-1 ... 1] (0 is 50:50 balanced)",
        "width": 5,
        "format": "5.2f",
        "align": ">",
    },
    "ulr": {
        "dict_key": "uptime_lifetime_ratio",
        "description": "ratio of uptime to lifetime of channel [0 ... 1]",
        "width": 5,
        "format": "5.2f",
        "align": ">",
    },
    "imed": {
        "dict_key": "median_forwarding_in",
        "description": "median forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "imean": {
        "dict_key": "mean_forwarding_in",
        "description": "mean forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "imax": {
        "dict_key": "largest_forwarding_amount_in",
        "description": "largest forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "omed": {
        "dict_key": "median_forwarding_out",
        "description": "median forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "omean": {
        "dict_key": "mean_forwarding_out",
        "description": "mean forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "omax": {
        "dict_key": "largest_forwarding_amount_out",
        "description": "largest forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "out": {
        "dict_key": "total_forwarding_out",
        "description": "total forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "sr/w": {
        "dict_key": "sent_received_per_week",
        "description": "sent and received per week [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
}

PRINT_PEERS_FORMAT = {
    "alias": {
        "dict_key": "alias",
        "description": "alias",
        "width": ALIAS_LENGTH,
        "format": ".<" + str(ALIAS_LENGTH),
        "align": "^",
        "convert": lambda x: alias_cutoff(x),
    },
    "cap": {
        "dict_key": "total_capacity",
        "description": "total capacity [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "mpc": {
        "dict_key": "max_public_capacity",
        "description": "maximum public channel capacity [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "nid": {
        "dict_key": "node_id",
        "description": "node id",
        "width": 66,
        "format": "",
        "align": "^",
    },
    "flow": {
        "dict_key": "flow_direction",
        "description": "flow direction (positive is outwards)",
        "width": 5,
        "format": "5.2f",
        "align": ">",
    },
    "fio": {
        "dict_key": "fees_in_out",
        "description": "total fees in and out [sat]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "fo/w": {
        "dict_key": "fees_out_per_week",
        "description": "total fees out per week [sat / week]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "fi/w": {
        "dict_key": "fees_in_per_week",
        "description": "total fees per in week [sat / week]",
        "width": 7,
        "format": "7.2f",
        "align": ">",
        "convert": lambda x: float(x) / 1000,
    },
    "nc": {
        "dict_key": "number_channels",
        "description": "number of channels",
        "width": 2,
        "format": "2d",
        "align": "<",
    },
    "np": {
        "dict_key": "number_private_channels",
        "description": "number of private channels",
        "width": 2,
        "format": "2d",
        "align": "<",
    },
    "na": {
        "dict_key": "number_active_channels",
        "description": "number of active channels",
        "width": 2,
        "format": "2d",
        "align": "<",
    },
    "mlb": {
        "dict_key": "max_local_balance",
        "description": "maximal local balance [sat]",
        "width": 8,
        "format": "8d",
        "align": ">",
    },
    "mrb": {
        "dict_key": "max_remote_balance",
        "description": "maximal remote balance [sat]",
        "width": 8,
        "format": "8d",
        "align": ">",
    },
    "lbf": {
        "dict_key": "local_base_fee",
        "description": "median local base fee [msat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lfr": {
        "dict_key": "local_fee_rate",
        "description": "median local fee rate",
        "width": 8,
        "format": "1.6f",
        "align": ">",
        "convert": lambda x: x / 1e6,
    },
    "lup": {
        "dict_key": "last_update",
        "description": "last update time [days ago]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lupp": {
        "dict_key": "last_update_peer",
        "description": "last update time by peer [days ago]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lupl": {
        "dict_key": "last_update_local",
        "description": "last update time by local [days ago]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "lb": {
        "dict_key": "local_balance",
        "description": "total local balance [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "nfwd": {
        "dict_key": "number_forwardings",
        "description": "number of forwardings",
        "width": 4,
        "format": "4.0f",
        "align": ">",
    },
    "priv": {
        "dict_key": "private",
        "description": "channel is private",
        "width": 5,
        "format": "^5",
        "align": ">",
        "convert": lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    "rb": {
        "dict_key": "remote_balance",
        "description": "total remote balance [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "r": {
        "dict_key": "routable",
        "description": "all channels are routable (active and at least one public channel)",
        "width": 1,
        "format": "^1",
        "align": ">",
        "convert": lambda x: NEGATIVE_MARKER if x else "",
    },
    "in": {
        "dict_key": "total_forwarding_in",
        "description": "total forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "ini": {
        "dict_key": "initiator",
        "description": "true if we opened channel",
        "width": 3,
        "format": "^3",
        "align": ">",
        "convert": lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    "tot": {
        "dict_key": "total_forwarding",
        "description": "total forwarding [sat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "ub": {
        "dict_key": "unbalancedness",
        "description": "unbalancedness [-1 ... 1] (0 is 50:50 balanced)",
        "width": 5,
        "format": "5.2f",
        "align": ">",
    },
    "ulr": {
        "dict_key": "uptime_lifetime_ratio",
        "description": "ratio of uptime to lifetime of channel [0 ... 1]",
        "width": 5,
        "format": "5.2f",
        "align": ">",
    },
    "imed": {
        "dict_key": "median_forwarding_in",
        "description": "median forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "imean": {
        "dict_key": "mean_forwarding_in",
        "description": "mean forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "imax": {
        "dict_key": "largest_forwarding_amount_in",
        "description": "largest forwarding inwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "omed": {
        "dict_key": "median_forwarding_out",
        "description": "median forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "omean": {
        "dict_key": "mean_forwarding_out",
        "description": "mean forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "omax": {
        "dict_key": "largest_forwarding_amount_out",
        "description": "largest forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "out": {
        "dict_key": "total_forwarding_out",
        "description": "total forwarding outwards [sat]",
        "width": 10,
        "format": "10.0f",
        "align": ">",
    },
    "sr/w": {
        "dict_key": "sent_received_per_week",
        "description": "sent and received per week [sat]",
        "width": 9,
        "format": "9d",
        "align": ">",
    },
    "rbf": {
        "dict_key": "remote_base_fee",
        "description": "median remote base fee [msat]",
        "width": 5,
        "format": "5.0f",
        "align": ">",
    },
    "rfr": {
        "dict_key": "remote_fee_rate",
        "description": "median remote fee rate",
        "width": 8,
        "format": "1.6f",
        "align": ">",
        "convert": lambda x: x / 1e6,
    },
}


def alias_cutoff(alias: str) -> str:
    """Cuts off the node alias at a certain length and removes unicode
    characters from the node alias."""

    if len(alias) > ALIAS_LENGTH:
        return alias[: ALIAS_LENGTH - 3] + "..."
    else:
        return alias


def _sorting_order(sort_string: str) -> Tuple[str, bool]:
    """Determines the sorting string and the sorting order.

    If sort_string starts with 'rev_', the sorting order is reversed."""

    reverse_sorting = True
    if sort_string[:4] == "rev_":
        reverse_sorting = False
        sort_string = sort_string[4:]

    sort_string = PRINT_CHANNELS_FORMAT[sort_string]["dict_key"]

    return sort_string, reverse_sorting


def _row_string(column_values: Dict, print_format: Dict, columns: List[str]):
    """Constructs the formatted row string for table printing."""

    string = ""
    for column in columns:
        format_string = print_format[column]["format"]
        conversion_function = print_format[column].get("convert", lambda x: x)
        value = column_values[print_format[column]["dict_key"]]
        converted_value = conversion_function(value)
        string += f"{converted_value:{format_string}} "

    return string


def _print_objects(objects: Dict, print_format: Dict, columns: str, sort_dict: Dict):
    """General purpose node/channel printing."""

    if not objects:
        logger.info(">>> Did not find any channels.")

    objects = OrderedDict(
        sorted(objects.items(), key=sort_dict["function"], reverse=sort_dict["reverse"])
    )

    logger.info("Sorting channels by %s.", sort_dict["string"])

    logger.info("-------- Description --------")
    columns = columns.split(",")
    for column in columns:
        logger.info("%-10s %s", column, print_format[column]["description"])
    logger.info("-----------------------------")
    # prepare the column header
    column_header = ""
    for column in columns:
        column_label = print_format[column]["align"]
        column_width = print_format[column]["width"]
        column_header += f"{column:{column_label}{column_width}} "

    # print the channel data
    for channel_number, (_, channel_data) in enumerate(objects.items()):
        if not channel_number % 20:
            logger.info(column_header)
        row = _row_string(channel_data, print_format, columns)
        logger.info(row)


class ListChannels(object):
    """A class to list lightning channels."""

    def __init__(self, node: "LndNode"):
        self.node = node

    def print_all_channels(self, sort_string="rev_alias"):
        """Prints all active and inactive channels."""
        channels = self._add_channel_annotations(self.node.get_all_channels())
        sort_string, reverse_sorting = _sorting_order(sort_string)
        sort_dict = {
            "function": lambda x: (
                x[1][PRINT_CHANNELS_FORMAT["priv"]["dict_key"]],
                x[1][sort_string],
            ),
            "string": sort_string,
            "reverse": reverse_sorting,
        }
        _print_objects(
            channels,
            PRINT_CHANNELS_FORMAT,
            columns="cid,priv,act,ub,cap,lb,rb,lbf,lfr,annotation,alias",
            sort_dict=sort_dict,
        )

    def print_channels_unbalanced(self, unbalancedness: float, sort_string="rev_ub"):
        """Prints unbalanced channels with |unbalancedness(channel)| > unbalancedness."""
        channels = self._add_channel_annotations(
            self.node.get_unbalanced_channels(unbalancedness)
        )
        sort_string, reverse_sorting = _sorting_order(sort_string)
        sort_dict = {
            "function": lambda x: x[1][sort_string],
            "string": sort_string,
            "reverse": reverse_sorting,
        }
        _print_objects(
            channels,
            PRINT_CHANNELS_FORMAT,
            columns="cid,ub,cap,lb,rb,pbf,pfr,annotation,alias",
            sort_dict=sort_dict,
        )

    def print_channels_inactive(self, sort_string="lupp"):
        """Prints all inactive channels."""
        channels = self._add_channel_annotations(self.node.get_inactive_channels())
        sort_string, reverse_sorting = _sorting_order(sort_string)
        sort_dict = {
            "function": lambda x: (-x[1]["private"], x[1][sort_string]),
            "string": sort_string,
            "reverse": reverse_sorting,
        }

        _print_objects(
            channels,
            PRINT_CHANNELS_FORMAT,
            columns="cid,lupp,ulr,priv,ini,age,ub,cap,lb,rb,sr/w,annotation,alias",
            sort_dict=sort_dict,
        )

    def print_channels_forwardings(
        self, time_interval_start: float, time_interval_end: float, sort_string: str
    ):
        """Prints forwarding statistics for each channel.

        :param time_interval_start: int
        :param time_interval_end: int
        :param sort_string: str
        """
        channels = get_channel_properties(
            self.node, time_interval_start, time_interval_end
        )
        channels = self._add_channel_annotations(channels)
        sort_string, reverse_sorting = _sorting_order(sort_string)
        sort_dict = {
            "function": lambda x: (
                float("inf") if math.isnan(x[1][sort_string]) else x[1][sort_string],
                x[1][PRINT_CHANNELS_FORMAT["nfwd"]["dict_key"]],
                x[1][PRINT_CHANNELS_FORMAT["ub"]["dict_key"]],
            ),
            "string": sort_string,
            "reverse": reverse_sorting,
        }
        _print_objects(
            channels,
            PRINT_CHANNELS_FORMAT,
            columns="cid,nfwd,age,fees,fo/w,fi/w,flow,ub,bwd,r,"
            "cap,pbf,pfr,annotation,alias",
            sort_dict=sort_dict,
        )

    def print_channels_hygiene(self, time_interval_start: float, sort_string: str):
        """Prints hygiene statistics for each channel."""
        time_interval_end = time.time()
        channels = get_channel_properties(
            self.node, time_interval_start, time_interval_end
        )
        channels = self._add_channel_annotations(channels)
        sort_string, reverse_sorting = _sorting_order(sort_string)
        sort_dict = {
            "function": lambda x: x[1][sort_string],
            "string": sort_string,
            "reverse": reverse_sorting,
        }
        _print_objects(
            channels,
            PRINT_CHANNELS_FORMAT,
            columns="cid,age,ini,nfwd/a,nfwd,fo/w,ulr,lb,cap,lfr,pfr,annotation,alias",
            sort_dict=sort_dict,
        )

    def _add_channel_annotations(self, channels: Dict) -> Dict:
        """Appends metadata to existing channel dicts from the configuration file."""
        if self.node.config:
            logger.debug("Adding annotations from file %s.", self.node.config_file)
        # mapping between the channel point and channel id
        channel_point_mapping = {
            k: v["channel_point"].split(":")[0] for k, v in channels.items()
        }
        # only read annotations if config file is given
        if self.node.config_file:
            config = settings.read_config(self.node.config_file)
            annotations = config["annotations"]
        else:
            annotations = {}
        channel_annotations_funding_id = {}
        channel_annotations_channel_id = {}

        for chan_id, annotation in annotations.items():
            if len(chan_id) == 18 and chan_id.isnumeric():
                # valid channel id
                channel_annotations_channel_id[int(chan_id)] = annotation
            elif len(chan_id) == 64 and chan_id.isalnum():
                # valid funding transaction id
                channel_annotations_funding_id[chan_id] = annotation
            else:
                raise ValueError(
                    "First part needs to be either a channel id or the "
                    "funding transaction id. \n"
                    "The funding transaction id can be found with "
                    "`lncli listchannels` under the channel point (the "
                    "characters before the colon)."
                )

        for channel_id, channel_values in channels.items():
            # get the annotation by channel id first
            annotation = channel_annotations_channel_id.get(channel_id, None)
            # if no channel annotation, try with funding id
            if annotation is None:
                annotation = channel_annotations_funding_id.get(
                    channel_point_mapping[channel_id], None
                )

            if annotation is not None:
                channels[channel_id]["annotation"] = annotation
            else:
                channels[channel_id]["annotation"] = ""

        return channels

    @staticmethod
    def _row_string(column_values, columns):
        """
        Constructs the formatted row string for table printing.

        :param column_values: dict
        :param columns: list of str
        :return: formatted str
        """

        string = ""
        for column in columns:
            format_string = PRINT_CHANNELS_FORMAT[column]["format"]
            conversion_function = PRINT_CHANNELS_FORMAT[column].get(
                "convert", lambda x: x
            )
            value = column_values[PRINT_CHANNELS_FORMAT[column]["dict_key"]]
            converted_value = conversion_function(value)
            string += f"{converted_value:{format_string}} "


class ListPeers(object):
    """A class to list lightning peers (with existing channels)."""

    def __init__(self, node: "LndNode"):
        self.node = node

    def print_all_nodes(
        self,
        time_interval_start: float,
        time_interval_end: float,
        sort_string: str = "fo/w",
    ):
        """Prints nodes with forwarding statistics."""

        nodes = get_node_properites(self.node, time_interval_start, time_interval_end)

        sort_string, reverse_sorting = _sorting_order(sort_string)
        sort_dict = {
            "function": lambda x: (
                x[1][sort_string],
                x[1][PRINT_PEERS_FORMAT["mpc"]["dict_key"]],
            ),
            "string": sort_string,
            "reverse": reverse_sorting,
        }

        _print_objects(
            nodes,
            PRINT_PEERS_FORMAT,
            columns="nid,alias,nc,na,np,nfwd,flow,ub,fo/w,fi/w,mpc,lb,rb,in,out,lfr,rfr",
            sort_dict=sort_dict,
        )


if __name__ == "__main__":
    import logging.config
    from lndmanage.lib.node import LndNode

    logging.config.dictConfig(settings.logger_config)
    logger = logging.getLogger()

    nd = LndNode("/home/user/.lndmanage/config.ini")
    lp = ListPeers(nd)
    lc = ListChannels(nd)
    lp.print_all_nodes(
        time_interval_start=time.time() - 3600 * 24 * 14,
        time_interval_end=time.time(),
        sort_string="fo/w",
    )



"""
Integration tests for rebalancing of channels.
"""
import time
from typing import Optional

from lndmanage import settings
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.ln_utilities import local_balance_to_unbalancedness
from lndmanage.lib.exceptions import NoRebalanceCandidates

from test.testing_common import (
    lndmanage_home,
    test_graphs_paths,
    SLEEP_SEC_AFTER_REBALANCING,
    TestNetwork
)

import logging.config
settings.set_lndmanage_home_dir(lndmanage_home)
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.handlers[0].setLevel(logging.DEBUG)


class RebalanceTest(TestNetwork):
    """
    Implements an abstract testing class for channel rebalancing.
    """
    def rebalance_and_check(
            self,
            test_channel_number: int,
            target: Optional[float],
            amount_sat: Optional[int],
            allow_uneconomic: bool,
            places: int = 5,
    ):
        """
        Test function for rebalancing to a specific target unbalancedness and
        asserts afterwards that the target was reached.

        :param test_channel_number: channel id
        :param target: unbalancedness target
        :param amount_sat: rebalancing amount
        :param allow_uneconomic: if uneconomic rebalancing should be allowed
        :param places: accuracy of the comparison between expected and tested values
        :type places: int
        """
        graph_before = self.testnet.assemble_graph()

        rebalancer = Rebalancer(
            self.lndnode,
            max_effective_fee_rate=5E-6,
            budget_sat=20,
            force=allow_uneconomic,
        )

        channel_id = self.testnet.channel_mapping[
            test_channel_number]['channel_id']

        fees_msat = rebalancer.rebalance(
            channel_id,
            dry=False,
            target=target,
            amount_sat=amount_sat,
        )

        # sleep a bit to let LNDs update their balances
        time.sleep(SLEEP_SEC_AFTER_REBALANCING)

        # check if graph has the desired channel balances
        graph_after = self.testnet.assemble_graph()

        channel_data_before = graph_before['A'][test_channel_number]
        channel_data_after = graph_after['A'][test_channel_number]
        amount_sent = channel_data_before['local_balance'] - channel_data_after['local_balance']

        channel_unbalancedness, _ = local_balance_to_unbalancedness(
            channel_data_after['local_balance'],
            channel_data_after['capacity'],
            channel_data_after['commit_fee'],
            channel_data_after['initiator']
        )

        if target is not None:
            self.assertAlmostEqual(
                target, channel_unbalancedness, places=places)

        elif amount_sat is not None:
            self.assertAlmostEqual(
                amount_sat, amount_sent, places=places)

        return fees_msat

    def graph_test(self):
        """
        graph_test should be implemented by each subclass test and check,
        whether the test graph has the correct shape.
        """
        raise NotImplementedError


class TestLiquidRebalance(RebalanceTest):
    network_definition = test_graphs_paths['star_ring_3_liquid']

    # channels:
    # A -> B (channel #1)
    # A -> C (channel #2)
    # A -> D (channel #6)
    # B -> C (channel #3)
    # B -> D (channel #4)
    # C -> D (channel #5)

    def graph_test(self):
        self.assertEqual(6, len(self.master_node_graph_view))

    def test_non_init_balanced(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, target=0.0, amount_sat=None, allow_uneconomic=True)

    def test_non_init_small_positive_target(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, target=0.2, amount_sat=None, allow_uneconomic=True)

    def test_non_init_max_target(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, target=1.0, amount_sat=None, allow_uneconomic=True)

    def test_non_init_negative_target(self):
        # this test should fail when unbalancing is not allowed, as it would
        # unbalance another channel if the full target would be accounted for
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, target=-0.2, amount_sat=None, allow_uneconomic=True)

    def test_non_init_fail_due_to_economic(self):
        # this test should fail when unbalancing is not allowed, as it would
        # unbalance another channel if the full target would be accounted for
        test_channel_number = 6
        self.assertRaises(
            NoRebalanceCandidates,
            self.rebalance_and_check, test_channel_number, target=-0.2, amount_sat=None, allow_uneconomic=False)

    def test_init_balanced(self):
        test_channel_number = 1
        self.rebalance_and_check(test_channel_number, target=0.0, amount_sat=None, allow_uneconomic=True, places=1)

    def test_init_already_balanced(self):
        test_channel_number = 2
        self.rebalance_and_check(test_channel_number, target=0.0, amount_sat=None, allow_uneconomic=True, places=2)

    def test_init_default_amount(self):
        test_channel_number = 1
        self.rebalance_and_check(test_channel_number, target=None, amount_sat=None, allow_uneconomic=True, places=-1)

    def test_shuffle_arround(self):
        """Shuffles sats around in channel 6."""
        first_target_amount = -0.1
        second_target_amount = 0.1
        test_channel_number = 6

        self.rebalance_and_check(
            test_channel_number, target=first_target_amount, amount_sat=None, allow_uneconomic=True)
        self.rebalance_and_check(
            test_channel_number, target=second_target_amount, amount_sat=None, allow_uneconomic=True)


class TestUnbalancedRebalance(RebalanceTest):
    network_definition = test_graphs_paths['star_ring_4_unbalanced']

    # channels:
    # A -> B (channel #1)
    # A -> C (channel #2)
    # A -> D (channel #3)
    # A -> E (channel #4)

    def graph_test(self):
        self.assertEqual(10, len(self.master_node_graph_view))

    def test_channel_1(self):
        """tests multiple rebalance of one channel"""
        test_channel_number = 1
        print(self.rebalance_and_check(
            test_channel_number, target=-0.05, amount_sat=None, allow_uneconomic=True, places=1))


class TestIlliquidRebalance(RebalanceTest):
    network_definition = test_graphs_paths['star_ring_4_illiquid']

    # channels:
    # A -> B (channel #1)
    # A -> C (channel #2)
    # A -> D (channel #3)
    # A -> E (channel #4)

    def graph_test(self):
        self.assertEqual(10, len(self.master_node_graph_view))

    def test_channel_1_splitting(self):
        """Tests multiple payment attempts with splitting."""
        test_channel_number = 1
        fees_msat = self.rebalance_and_check(
            test_channel_number, target=-0.05, amount_sat=None, allow_uneconomic=True, places=1)
        self.assertAlmostEqual(2000, fees_msat, places=-3)




import os
import configparser
from lndmanage.lib.configure import check_or_create_configuration
from pathlib import Path

# -------- graph settings --------
# accepted age of the network graph
CACHING_RETENTION_MINUTES = 30

# -------- pathfinding --------
# default penalty for non-active channels / too small channels / unbalanced channels
PENALTY = 1E9
# if a penalty should be applied for long paths
PREFER_SHORT_PATHS = True
# penalty per hop in msat, used together with PREFER_SHORT_PATHS
LONG_PATH_PENALTY_MSAT = 2000
# exclude channels, which have less than amt * MIN_REL_CHANNEL_CAPACITY
# this can be used to increase success rates
MIN_REL_CHANNEL_CAPACITY = 0.75

# -------- network analysis --------
# a user typically has a low number of channels, this number
NUMBER_CHANNELS_DEFINING_USER_NODE = 3
NUMBER_CHANNELS_DEFINING_HUB = 200

# -------- rebalancing --------
# in terms of unbalancedness
UNBALANCED_CHANNEL = 0.2
# rebalancing will be done with CHUNK_SIZE of the minimal capacity
# of the to be balanced channels
REBALANCING_TRIALS = 10


logger_config = None
home_dir = None


def set_lndmanage_home_dir(directory=None):
    """
    Sets the correct path to the lndmanage home folder.

    :param directory: home folder, overwrites default
    :type directory: str
    """
    global home_dir, logger_config

    if directory:
        home_dir = directory
    else:
        # determine home folder, prioritized by environment
        # variable LNDMANAGE_HOME
        environ_home = os.environ.get('LNDMANAGE_HOME')

        if environ_home:
            if not os.path.isabs(environ_home):
                raise ValueError(
                    f'Environment variable LNDMANAGE_HOME must be '
                    f'an absolute path. Current: "{environ_home}"')
            home_dir = environ_home
        else:
            user_home_dir = str(Path.home())
            home_dir = os.path.join(user_home_dir, '.lndmanage')

        # if lndmanage is ran for the first time,
        # we need to create the configuration
        check_or_create_configuration(home_dir)

    # logger settings
    logfile_path = os.path.join(home_dir, 'lndmanage.log')

    logger_config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'file': {
                'format': '[%(asctime)s %(levelname)s] %(message)s',
                #'format': '[%(asctime)s %(levelname)s %(name)s] %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S'
            },
            'standard': {
                'format': '%(message)s',
            },
        },
        'handlers': {
            'default': {
                'level': 'INFO',
                'formatter': 'standard',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',  # Default is stderr
            },
            'file': {
                'level': 'DEBUG',
                'formatter': 'file',
                'class': 'logging.FileHandler',
                'filename': logfile_path,
                'encoding': 'utf-8',
            },
        },
        'loggers': {
            '': {  # root logger
                'handlers': ['default', 'file'],
                'level': 'DEBUG',
                'propagate': True
            },
        }
    }


def read_config(config_path):
    config = configparser.ConfigParser()
    config.read(config_path)
    return config


set_lndmanage_home_dir()



"""
Handling lncli interaction.
"""
import os
import subprocess
import json

from pygments import highlight, lexers, formatters
from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Lncli(object):
    def __init__(self, lncli_path, config_file):
        self.lncli_path = lncli_path

        config = settings.read_config(config_file)

        cert_file = os.path.expanduser(config['network']['tls_cert_file'])
        macaroon_file = \
            os.path.expanduser(config['network']['admin_macaroon_file'])
        lnd_host = config['network']['lnd_grpc_host']

        # assemble the command for lncli for execution with flags
        self.lncli_command = [
            self.lncli_path,
            '--rpcserver=' + lnd_host,
            '--macaroonpath=' + macaroon_file,
            '--tlscertpath=' + cert_file
            ]

    def lncli(self, command):
        """
        Invokes the lncli command line interface for lnd.

        :param command: list of command line arguments
        :return:
            int: error code
        """

        cmd = self.lncli_command + command
        logger.debug('executing lncli %s', ' '.join(cmd))
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # check if the output can be decoded from valid json
        try:
            json.loads(proc.stdout)
            # convert json into color coded characters
            colorful_json = highlight(
                proc.stdout,
                lexers.JsonLexer(),
                formatters.TerminalFormatter()
            )
            logger.info(colorful_json)

        # usually errors and help are not json, handle them here
        except ValueError:
            logger.info(proc.stdout.decode('utf-8'))
            logger.info(proc.stderr.decode('utf-8'))

        return proc.returncode



from typing import List, Dict, TYPE_CHECKING, Tuple

from lndmanage.lib.exceptions import RouteWithTooSmallCapacity, NoRoute
from lndmanage.lib.pathfinding import dijkstra
from lndmanage import settings

if TYPE_CHECKING:
    from lndmanage.lib.node import LndNode

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def calculate_fees_on_policy(amt_msat, policy):
    return policy['fee_base_msat'] + amt_msat * policy['fee_rate_milli_msat'] // 1000000


class Route(object):
    """Deals with the onion route construction from list of channels. Calculates fees
    and cltvs.
    """

    def __init__(self, node: 'LndNode', channel_hops: List[int], node_dest: str, amt_msat: int):
        """
        :param node: :class:`lib.node.Node` instance
        :param channel_hops: list of chan_ids along which the route shall be constructed
        :param node_dest: pub_key of destination node
        :param amt_msat: amount to send in msat
        """

        self.node = node
        self.blockheight = node.blockheight
        logger.debug(f"Blockheight: {self.blockheight}")
        self.channel_hops = channel_hops
        self._hops = []
        self._node_hops = [node_dest]
        forward_msat = amt_msat
        final_cltv = 144
        fees_msat_container = [0]
        cltv_delta = [0]
        node_to = node_dest
        node_from = None
        policy = None

        logger.debug("Route construction starting.")
        # hops are traversed in backwards direction to accumulate fees and cltvs
        for ichannel, channel_id in enumerate(reversed(channel_hops)):
            channel_data = self.node.network.edges[channel_id]
            # TODO: add private channels for sending

            if amt_msat // 1000 > channel_data['capacity']:
                logger.debug(f"Discovered a channel {channel_id} with too small capacity.")
                raise RouteWithTooSmallCapacity(f"Amount too large for channel.")

            policies = channel_data['policies']
            if node_to == channel_data['node2_pub']:
                try:
                    policy = policies[channel_data['node1_pub'] > channel_data['node2_pub']]
                    node_from = channel_data['node1_pub']
                except KeyError:
                    logger.exception(f"No channel {channel_data}")
            else:
                policy = policies[channel_data['node2_pub'] > channel_data['node1_pub']]
                node_from = channel_data['node2_pub']

            self._node_hops.append(node_from)
            hop = len(channel_hops) - ichannel
            logger.info(f"    Hop {hop}: {channel_id} (cap: {channel_data['capacity']} sat): "
                        f"{self.node.network.node_alias(node_to)} <- {self.node.network.node_alias(node_from)} ")
            logger.debug(f"      Policy of forwarding node: {policy}")

            fees_msat = policy['fee_base_msat'] + policy['fee_rate_milli_msat'] * forward_msat // 1000000
            forward_msat = amt_msat + sum(fees_msat_container[:ichannel])
            fees_msat_container.append(fees_msat)

            logger.info(f"      Fees: {fees_msat / 1000 if not hop == 1 else 0:3.3f} sat")
            logger.debug(f"      Fees container {fees_msat_container}")
            logger.debug(f"      Forward: {forward_msat / 1000:3.3f} sat")
            logger.info(f"      Liquidity penalty: {self.node.network.liquidity_hints.penalty(node_from, node_to, channel_data, amt_msat, self.node.network.channel_rater.reference_fee_rate_milli_msat) / 1000: 3.3f} sat")
            logger.info(f"      Badness penalty: {self.node.network.liquidity_hints.badness_penalty(node_from, amt_msat) / 1000: 3.3f} sat")
            logger.info(f"      Time penalty: {self.node.network.liquidity_hints.time_penalty(node_from, amt_msat) / 1000: 3.3f} sat")

            self._hops.append({
                'chan_id': channel_data['channel_id'],
                'chan_capacity': channel_data['capacity'],
                'amt_to_forward': forward_msat // 1000,
                'fee': fees_msat_container[-2] // 1000,
                'expiry': self.blockheight + final_cltv + sum(cltv_delta[:ichannel]),
                'amt_to_forward_msat': forward_msat,
                'fee_msat': fees_msat_container[-2],
            })

            cltv_delta.append(policy['time_lock_delta'])
            node_to = node_from

        self.hops = list(reversed(self._hops))
        self.node_hops = list(reversed(self._node_hops))
        self.total_amt_msat = amt_msat + sum(fees_msat_container[:-1])
        self.total_fee_msat = sum(fees_msat_container[:-1])
        self.total_time_lock = sum(cltv_delta[:-1]) + self.blockheight + final_cltv

    def _debug_route(self):
        """Prints detailed information of the route."""
        logger.debug("Debug route:")
        for h in self.hops:
            logger.debug(f"c:{h['chan_id']} a:{h['amt_to_forward']} f:{h['fee_msat']}"
                         f" c:{h['expiry'] - self.node.blockheight}")
            channel_info = self.node.network.edges[h['chan_id']]
            node1 = channel_info['node1_pub']
            node2 = channel_info['node2_pub']
            logger.debug(f"{node1[:5]}, {channel_info['node1_policy']}")
            logger.debug(f"{node2[:5]}, {channel_info['node2_policy']}")
        logger.debug(f"tl:{self.total_time_lock} ta:{self.total_amt_msat} tf:{self.total_fee_msat}")


class Router(object):
    """Contains utilities for constructing routes."""

    def __init__(self, node: 'LndNode'):
        self.node = node

    def _node_route_to_channel_route(self, node_route: List[str], amt_msat: int) -> List[int]:
        """Takes a route in terms of a list of nodes and translates it into a list of
        channels.

        :param node_route: list of pubkeys
        :param amt_msat: amount to send in sat
        :return: list of channel_ids
        """
        channels = []
        for p in range(len(node_route) - 1):
            channels.append(
                self._determine_channel(
                    node_route[p], node_route[p + 1], amt_msat)[1])
        return channels

    def get_route_from_to_nodes(self, node_from: str, node_to: str, amt_msat: int) -> List[str]:
        """Determines number_of_routes shortest paths between node_from and node_to for
        an amount of amt_msat.

        :param node_from: pubkey
        :param node_to: pubkey
        :param amt_msat: amount to send in msat
        :return: route
        """
        self.node.network.channel_rater.blacklisted_nodes.append(self.node.pub_key)  # excludes self-loops

        weight_function = lambda v, u, e: self.node.network.channel_rater.node_to_node_weight(v, u, e, amt_msat)
        route = dijkstra(self.node.network.graph, node_from, node_to, weight=weight_function)

        if not route:
            raise NoRoute

        return route

    def _determine_channel(self, node_from: str, node_to: str, amt_msat: int):
        """Determines the cheapest channel between nodes node_from and node_to for an
        amount of amt_msat.

        :param node_from: pubkey
        :param node_to: pubkey
        :param amt_msat: amount to send in msat
        :return: channel_id
        """
        number_edges = self.node.network.graph.number_of_edges(node_from, node_to)
        channels_with_calculated_fees = []
        for n in range(number_edges):
            edge = self.node.network.graph.get_edge_data(node_from, node_to, n)
            fees = self.node.network.channel_rater.channel_weight(node_from, node_to, edge, amt_msat)
            channels_with_calculated_fees.append([fees, edge['channel_id']])
        sorted_channels = sorted(channels_with_calculated_fees, key=lambda k: k[0])
        best_channel = sorted_channels[0]
        # we check that we don't encounter a hop which is blacklisted
        if best_channel[0] == float('inf'):
            raise NoRoute('channels graph exhausted')
        return best_channel

    def _determine_cheapest_fees_between_two_nodes(self, node_from, node_to, amt_msat):
        return self._determine_channel(node_from, node_to, amt_msat)[0]

    def get_route_channel_hops_from_to_node_internal(
            self,
            source_pubkey: str,
            target_pubkey: str,
            amt_msat: int
    ) -> List[int]:
        """Find routes internally, using networkx to construct a route from a source
        node to a target node."""
        logger.debug(f"Internal pathfinding:")
        logger.debug(f"from {source_pubkey}")
        logger.debug(f"  to {target_pubkey}")

        node_route = self.get_route_from_to_nodes(
            source_pubkey, target_pubkey, amt_msat)

        return self._node_route_to_channel_route(node_route, amt_msat)

    def get_route(
            self,
            send_channels: Dict[int, dict],
            receive_channels: Dict[int, dict],
            amt_msat: int
    ) -> Route:
        """Calculates a route from send_channels to receive_channels.

        :param send_channels: channel ids to send from
        :param receive_channels: channel ids to receive to
        :param amt_msat: payment amount in msat

        :return: a route for rebalancing
        """
        this_node = self.node.pub_key # TODO: make this a parameter for general route calculation
        self.node.network.channel_rater.reset_channel_blacklist()

        # We will ask for a route from source to target.
        # we send via a send channel and receive over other channels:
        # this_node -(send channel)-> source -> ... -> receiver neighbors -(receive channels)-> target (this_node)
        if len(send_channels) == 1:
            send_channel = list(send_channels.values())[0]
            source = send_channel['remote_pubkey']
            target = this_node

            # we don't want to go backwards via the send_channel (and other parallel channels)
            channels_source_target = self.node.network.graph[source][target]
            for channel in channels_source_target.values():
                self.node.network.channel_rater.blacklist_add_channel(channel['channel_id'], source, target)

            # we want to use the receive channels for receiving only, so don't receive over other channels
            excluded_receive_channels = self.node.get_unbalanced_channels(
                excluded_channels=[k for k in receive_channels.keys()], public_only=False, active_only=False)
            for channel_id, channel in excluded_receive_channels.items():
                receiver_neighbor = channel['remote_pubkey']
                self.node.network.channel_rater.blacklist_add_channel(channel_id, receiver_neighbor, target)

        # we send via several channels and receive over a single one:
        # this_node (source) -(send channels)-> ... -> receiver neighbor (target) -(receive channel)-> this_node
        elif len(receive_channels) == 1:
            receive_channel = list(receive_channels.values())[0]
            source = this_node
            target = receive_channel['remote_pubkey']

            # we want to block the receiving channel (and parallel ones) from sending
            channels_source_target = self.node.network.graph[source][target]
            for channel in channels_source_target.values():
                self.node.network.channel_rater.blacklist_add_channel(channel['channel_id'], source, target)

            # we want to use the send channels for sending only, so don't send over other channels
            excluded_send_channels = self.node.get_unbalanced_channels(
                excluded_channels=[k for k in send_channels.keys()], public_only=False, active_only=False)
            for channel_id, channel in excluded_send_channels.items():
                sender_neighbor = channel['remote_pubkey']
                self.node.network.channel_rater.blacklist_add_channel(channel_id, source, sender_neighbor)
        else:
            raise ValueError("One of the two channel sets should be singular.")

        # determine inner channel hops
        # internal method uses networkx dijkstra,
        # this is more independent, but slower
        route_channel_hops = \
            self.get_route_channel_hops_from_to_node_internal(
            source, target, amt_msat)

        final_channel_hops = []
        if len(send_channels) == 1:
            final_channel_hops.append(send_channel['chan_id'])
            final_channel_hops.extend(route_channel_hops)
        else:
            final_channel_hops.extend(route_channel_hops)
            final_channel_hops.append(receive_channel['chan_id'])

        # TODO: add some consistency checks, route shouldn't contain self-loops
        logger.debug("Channel hops:")
        logger.debug(final_channel_hops)

        # initialize Route objects with appropriate fees and expiries
        route = Route(self.node, final_channel_hops, this_node, amt_msat)

        return route


if __name__ == '__main__':
    import logging.config
    logging.config.dictConfig(settings.logger_config)
    from lndmanage.lib.node import LndNode
    nd = LndNode()



"""
Implements a lightning network topology with only one node and no channels.
"""
nodes = {
    'A': {
        'grpc_port': 11009,
        'rest_port': 8080,
        'port': 9735,
        'base_fee_msat': 1,
        'fee_rate': 0.000001,
        'channels': {
        }
    },
}




import datetime
from typing import Dict

from lndmanage.lib.network_info import NetworkAnalysis
from lndmanage.lib.ln_utilities import (
    parse_nodeid_channelid,
    convert_channel_id_to_short_channel_id,
    height_to_timestamp
)
from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# width of a column in the ouptput in characters
COL_WIDTH = 66


def padded_column_string(string1, string2, unit, shift=10, max_len_string1=25):
    """
    Padded_column_string is a helper function which returns a formatted string
    representing object, quantity and unit of a piece of information.

    :param string1: information
    :type string1: str
    :param string2: quantity
    :type string2: str
    :param unit: unit of the information
    :type unit: str
    :param shift: whitespace shift to the right (indentation)
    :type shift: int
    :param max_len_string1: needed to calculate padding to the right
    :type max_len_string1: int
    :return:
    :rtype:
    """

    string = f" " * shift + f"{string1:<{max_len_string1}} {string2} {unit}"
    if type(string2) == float:
        string = f" " * shift + f"{string1:<{max_len_string1}} " \
            f"{string2:0.6f} {unit}"
    return string


class Info(object):
    """
    Implements the info command, which displays info on individual channels and
    nodes.
    """
    def __init__(self, node):
        """
        :param node: node object
        :type node: lndmanage.lib.node.LndNode
        """
        self.node = node
        self.network_info = NetworkAnalysis(self.node)

    def parse_and_print(self, info):
        """
        Parses an info string for a channel id or node public key and prints
        out the information gathered about the object.

        :param info: channel id or node public key
        :type info: str
        """

        # analyzer = NetworkAnalysis(self.node)
        try:
            channel_id, node_pub_key = parse_nodeid_channelid(info)
        except ValueError:
            logger.info("Info didn't represent neither a channel nor a node.")
            return

        # Info was a channel.
        if channel_id is not None:
            try:
                general_info = self.node.network.edges[channel_id]
            except KeyError:
                logger.info("Channel id %s is not known in the public graph.",
                            channel_id)
                return

            # Add some more information on the channel.
            general_info['node1_alias'] = \
                self.node.network.node_alias(general_info['node1_pub'])
            general_info['node2_alias'] = \
                self.node.network.node_alias(general_info['node2_pub'])
            general_info['blockheight'] = \
                convert_channel_id_to_short_channel_id(
                    channel_id)[0]
            general_info['open_timestamp'] = height_to_timestamp(
                self.node, general_info['blockheight'])

            # TODO: if it's our channel, add extra info
            extra_info = None

            self.print_channel_info(general_info)

        # Info was a node.
        else:
            try:
                general_info = self.network_info.node_info_basic(node_pub_key)
            except KeyError:
                return

            # TODO: if it's a (channel) peer or our node, add extra info
            extra_info = None

            self.print_node_info(general_info)

    def print_channel_info(self, general_info: Dict):
        """
        Prints the channel info with peer information.

        :param general_info: information about the channel in the public graph
        """

        logger.info("-------- Channel info --------")
        logger.info(f"channel id: {general_info['channel_id']}  "
                    f"channel point: {general_info['chan_point']}")

        # capactiy
        string = padded_column_string(
            'capacity:', general_info['capacity'], 'sat')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # blockheight
        string = padded_column_string(
            'blockheight:', general_info['blockheight'], '')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # opening time
        time = datetime.datetime.utcfromtimestamp(
            general_info['open_timestamp']).strftime('%Y-%m-%d %H:%M:%S')
        string = padded_column_string('open since:', time, '')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # channel age
        age = round(
            (self.node.blockheight - general_info['blockheight']) / 6 / 24, 2)
        string = padded_column_string('channel age:', age, 'days')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # last update
        last_update = general_info['last_update']
        last_update_time = datetime.datetime.utcfromtimestamp(
            last_update).strftime('%Y-%m-%d %H:%M:%S')
        string = padded_column_string('last update:', last_update_time, '')
        logger.info(f"{string:{COL_WIDTH*2}}")
        logger.info("")

        # channel partner overview
        logger.info("-------- Channel partners --------")
        logger.info(f"{general_info['node1_pub']:{COL_WIDTH}} | "
                    f"{general_info['node2_pub']:{COL_WIDTH}}")
        logger.info(f"{general_info['node1_alias']:^{COL_WIDTH}} | "
                    f"{general_info['node2_alias']:^{COL_WIDTH}}")

        policies = general_info['policies']
        np1 = policies[general_info['node1_pub'] > general_info['node2_pub']]
        np2 = policies[general_info['node2_pub'] > general_info['node1_pub']]
        last_update_1 = np1['last_update']
        last_update_2 = np2['last_update']

        last_update_time_1 = datetime.datetime.utcfromtimestamp(
            last_update_1).strftime('%Y-%m-%d %H:%M:%S')
        last_update_time_2 = datetime.datetime.utcfromtimestamp(
            last_update_2).strftime('%Y-%m-%d %H:%M:%S')

        # base fee
        string_left = padded_column_string(
            'base fee:', np1['fee_base_msat'], 'msat')
        string_right = padded_column_string(
            'base fee:', np2['fee_base_msat'], 'msat')
        logger.info(
            f"{string_left:{COL_WIDTH}} | {string_right:{COL_WIDTH}}")

        # fee rate
        string_left = padded_column_string(
            'fee rate:', np1['fee_rate_milli_msat'] / 1E6, 'sat/sat')
        string_right = padded_column_string(
            'fee rate:', np2['fee_rate_milli_msat'] / 1E6, 'sat/sat')
        logger.info(
            f"{string_left:{COL_WIDTH}} | {string_right:{COL_WIDTH}}")

        # time lock delta
        string_left = padded_column_string(
            'time lock delta:', np1['time_lock_delta'], 'blocks')
        string_right = padded_column_string(
            'time lock delta:', np2['time_lock_delta'], 'blocks')
        logger.info(
            f"{string_left:{COL_WIDTH}} | {string_right:{COL_WIDTH}}")

        # disabled
        string_left = padded_column_string('disabled:', np1['disabled'], '')
        string_right = padded_column_string('disabled:', np2['disabled'], '')
        logger.info(
            f"{string_left:{COL_WIDTH}} | {string_right:{COL_WIDTH}}")

        # last update
        string_left = padded_column_string(
            'last update:', last_update_time_1, '')
        string_right = padded_column_string(
            'last update:', last_update_time_2, '')
        logger.info(
            f"{string_left:{COL_WIDTH}} | {string_right:{COL_WIDTH}}")

    def print_node_info(self, general_info):
        """
        Prints the node info.

        :param general_info: information about the node in the public graph
        :type general_info: dict
        """
        logger.info("-------- Node info --------")
        logger.info(general_info['pub_key'])

        # alias
        string = padded_column_string('alias:', general_info['alias'], '')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # last update
        last_update = general_info['last_update']
        last_update_time = datetime.datetime.utcfromtimestamp(
            last_update).strftime('%Y-%m-%d %H:%M:%S')
        string = padded_column_string('last update:', last_update_time, '')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # numer of channels
        string = padded_column_string(
            'number of channels:', general_info['num_channels'], '')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # total capacity
        string = padded_column_string(
            'total capacity:', general_info['total_capacity'], 'sat')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # capacity per channel
        string = padded_column_string(
            'capacity (median):', general_info['median_capacity'], 'sat')
        logger.info(f"{string:{COL_WIDTH*2}}")
        string = padded_column_string(
            'capacity (mean):', general_info['mean_capacity'], 'sat')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # fees
        string = padded_column_string(
            'base fee (median):', general_info['median_base_fee'], 'msat')
        logger.info(f"{string:{COL_WIDTH*2}}")
        string = padded_column_string(
            'base fee (mean):', general_info['mean_base_fee'], 'msat')
        logger.info(f"{string:{COL_WIDTH*2}}")
        string = padded_column_string(
            'fee rate (median):', general_info['median_fee_rate'], 'sat/sat')
        logger.info(f"{string:{COL_WIDTH*2}}")
        string = padded_column_string(
            'fee rate (mean):', general_info['mean_fee_rate'], 'sat/sat')
        logger.info(f"{string:{COL_WIDTH*2}}")

        # addresses
        logger.info("-------- Addresses --------")
        for addr in general_info['addresses']:
            logger.info(5 * " " + general_info['pub_key'] + "@" + addr)



import math
from typing import TYPE_CHECKING, Dict

from lndmanage import settings

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if TYPE_CHECKING:
    from lndmanage.lib.network import Network

BADNESS_RATE = 0.000_100


def node_badness(node_number: int, hop: int):
    return BADNESS_RATE * math.exp(-abs(node_number - (hop + 0.5)))


class ChannelRater:
    """The purpose of this class is to hold information about the balancedness of
    channels. It also defines the cost function of a channel with the function
    :func:`ChannelRater.lightning_channel_weight`.
    """

    def __init__(self, network: "Network", source: str = None):
        self.blacklisted_channels = {}
        self.blacklisted_nodes = []
        self.source = source
        self.network = network
        self.reference_fee_rate_milli_msat = 200

    def node_to_node_weight(self, u, v, e, amt_msat):
        """Is used to assign a weight for a channel. It is calculated from the fee
        policy of that channel:
        base fee + proportional fee + path penalty + capacity penalty + blacklist penalty

        :param u: source node
        :param v: target node
        :param e: fee policy of the edges
        :param amt_msat: amount in msat
        :return: cost of the channel in msat
        """
        node_penalty = 0
        if u in self.blacklisted_nodes or v in self.blacklisted_nodes:
            node_penalty = settings.PENALTY

        costs = [
            node_penalty + self.channel_weight(u, v, edge_properties, amt_msat)
            for edge_properties in e.values()
        ]
        return min(costs)

    def channel_weight(self, node_from: str, node_to: str, channel_info: Dict, amt_msat):
        # check if channel is blacklisted
        if self.blacklisted_channels.get(channel_info["channel_id"]) == {"source": node_from, "target": node_to}:
            return math.inf

        # we don't send if the channel cannot carry the payment
        if amt_msat // 1000 > channel_info["capacity"]:
            return math.inf

        # we don't send if the minimal htlc amount is not respected
        if amt_msat < channel_info.get("fees")[node_from > node_to]['min_htlc']:
            return math.inf

        # we don't send if the max_htlc_msat is not respected
        if amt_msat > channel_info.get("fees")[node_from > node_to]['max_htlc_msat']:
            return math.inf

        # we don't send over channel if it is disabled
        policy = channel_info.get("fees")[node_from > node_to]
        if policy["disabled"]:
            return math.inf

        # we don't pay fees if we own the channel and are sending over it
        if self.source and node_from == self.source:
            return 0

        # we apply a badness score proportional to the amount we send
        badness_penalty = self.network.liquidity_hints.badness_penalty(node_from, amt_msat)

        # compute liquidity penalty
        liquidity_penalty = self.network.liquidity_hints.penalty(
            node_from, node_to, channel_info, amt_msat, self.reference_fee_rate_milli_msat
        )

        # routing fees
        fees = (
            abs(policy["fee_base_msat"])
            + amt_msat * abs(policy["fee_rate_milli_msat"]) // 1_000_000
        )

        # we give a base penalty of 2 sat per hop, as we want to avoid too long routes
        # for small amounts
        route_length_fee_msat = 2_000

        # we discount on badness if we know that the channel can route
        if liquidity_penalty == 0:
            badness_penalty /= 2

        # time penalty
        time_penalty = self.network.liquidity_hints.time_penalty(node_from, amt_msat)

        # linear combination of components
        weight = fees + liquidity_penalty + badness_penalty + route_length_fee_msat + time_penalty

        return weight

    def blacklist_add_channel(self, channel: int, source: str, target: str):
        """Adds a channel to the blacklist dict.

        :param channel: channel_id
        :param source: pubkey
        :param target: pubkey
        """
        self.blacklisted_channels[channel] = {
            "source": source,
            "target": target,
        }

    def reset_channel_blacklist(self):
        self.blacklisted_channels = {}

    def blacklist_add_node(self, node_pub_key):
        """Adds a node public key to the blacklist.

        :param node_pub_key: str
        """
        self.blacklisted_nodes.append(node_pub_key)



from typing import Callable, List

import networkx as nx

from lndmanage.lib.utilities import profiled
from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@profiled
def dijkstra(graph: nx.Graph, source: str, target: str, weight: Callable) -> List[str]:
    """Wrapper for calculating a shortest path given a weight function.

    :param graph: networkx graph
    :param source: find a path from this key
    :param target: to this key
    :param weight: weight function, takes node_from (pubkey from), node_to (pubkey to), channel_info (edge information) as arguments

    :return: hops in terms of the node keys
    """
    path = nx.shortest_path(graph, source, target, weight=weight)
    return path



from configparser import NoSectionError
import json
import logging
import os
import time
from typing import Tuple, List, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from lndmanage.lib.node import LndNode

from lndmanage.lib.user import yes_no_question
from lndmanage.lib.forwardings import ForwardingAnalyzer
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

optimization_parameters = {
    "cltv": 40,  # blocks
    "min_base_fee": 0,  # msat
    "max_base_fee": 5000,  # msat
    "min_fee_rate": 0.000005,
    "max_fee_rate": 0.005000,
    "delta_max": 1.5,
    "delta_min_up": 0.05,
    "delta_min_dn": 0.50,
    "r_t": 100000 / 7,  # sat / day  TODO: optimize by statistical analysis
    "local_balance_reserve": 500000,  # sat
    "delta_b_min": 0.25,
    "delta_b_max": 0.50,
    "delta_b": 0.5,
    "n_t": 4 / 7,
}


def delta_min(params: dict, local_balance: int, capacity: int):
    """The capping from below for the delta_demand function."""
    if not local_balance <= capacity:
        raise ValueError(
            f"local balance must be lower than capacity "
            f"{local_balance} / {capacity}"
        )

    # if we have small channels, which can't respect the reserve, lower the
    # reserve
    if params["local_balance_reserve"] > capacity // 2:
        reserve = capacity // 3
    else:
        reserve = params["local_balance_reserve"]

    # if local balance is below balance reserve, start to charge more fees
    if local_balance < params["local_balance_reserve"]:
        x = params["delta_min_up"] / reserve
        return -x * (local_balance - reserve) + 1

    # if local balance is above balance reserve, charge less fees
    else:
        x = params["delta_min_dn"] / (capacity - reserve)
        return -x * (local_balance - reserve) + 1


def delta_demand(
    params: dict,
    time_interval: float,
    amount_out: float,
    local_balance: int,
    capacity: int,
) -> float:
    """Calculates a change factor for a channel by taking into account
    the amount transacted in a time interval compared to a target rate.

    The higher the amount forwarded, the larger the fee rate should be.
    :param params: fee optimization parameters
    :param time_interval: time interval in days
    :param amount_out: amount transacted outwards for the channel in sat
    :param local_balance: local balance in sat
    :param capacity: capacity in sat
    :return: demand adjustment factor"""
    r = amount_out / time_interval
    r_t = params["r_t"]

    logger.info(
        f"    Outward forwarded amount: {amount_out:6.0f} "
        f"(rate {r:5.0f} / target rate {r_t:5.0f})"
    )

    m = params["delta_min_dn"]

    c = 1.0 + m * (r / r_t - 1.0)

    mc = delta_min(params, local_balance, capacity)

    # cap from below
    if c < mc:
        return mc
    # cap from above
    elif c > params["delta_max"]:
        return params["delta_max"]
    else:
        return c


class FeeSetter(object):
    """Class for fee optimization."""

    def __init__(
        self, node: "LndNode", from_days_ago=7, parameters: Optional[dict] = None
    ):
        """
        :param node: node instance
        :param from_days_ago: forwarding history is taken over the past
            from_days_ago days
        :param parameters: fee algo parameters"""

        # by default, channel fees are updated, not initialized
        self.node = node

        self.history_path = os.path.join(settings.home_dir, "fee_history.log")

        if parameters is None:
            self.params = optimization_parameters
        else:
            self.params = parameters

        # initialize fee setter
        self.forwarding_analyzer = ForwardingAnalyzer(node)
        self.channel_fee_policies = node.get_channel_fee_policies()

        # determine time interval for forwardings analyzer to look back
        # for transactions
        self.time_end = time.time()
        self.time_start = self.time_end - from_days_ago * 24 * 60 * 60
        self.time_interval_days = from_days_ago
        self.forwarding_analyzer.initialize_forwarding_stats(
            self.time_start, self.time_end
        )

        # get channel info
        self.channels = self.node.get_all_channels()
        self.channels_forwarding_stats = (
            self.forwarding_analyzer.get_forwarding_statistics_channels()
        )

    def set_fees(self, init=False, reckless=False) -> List[dict]:
        """Sets channel fee policies considering different metrics like
        unbalancedness and demand.

        :param init: true if fees are set initially with this method
        :param reckless: if set, there won't be any user interaction
        :return: fee changes statistics"""

        channel_fee_policies, stats = self.new_fee_policies(init)

        if reckless:
            set_fees = True
        else:
            logger.info("Do you want to set these fees? Enter [yes/no]:")
            set_fees = yes_no_question()

        if set_fees:
            self.node.set_channel_fee_policies(channel_fee_policies)
            self.append_to_history(stats)
            logger.info("Have set new fee policy.")
        else:
            logger.info("Didn't set new fee policy.")

        return stats

    def new_fee_policies(self, init=False) -> Tuple[dict, list]:
        """Calculates and reports the changes to the new fee policy.

        :param init: when true, fee policy is initialized
        :return: (new channel policies, fee update statistics)"""
        logger.info("Determining new channel policies based on demand.")
        logger.info(
            "Every channel will have a base fee of %d msat and cltv " "of %d.",
            self.params["min_base_fee"],
            self.params["cltv"],
        )
        channel_fee_policies = {}

        try:
            ignored_channels = self.node.config.items("excluded-channels-fee-opt")
            ignored_channels = {int(c) for c, _ in ignored_channels}
        except NoSectionError:
            ignored_channels = set()

        stats = []

        # loop over channel peers
        for pk, cs in self.node.pubkey_to_channel_map().items():
            ignore_peer = bool(set(cs).intersection(ignored_channels))
            logger.info(
                f">>> Fee optimization for node {pk} "
                f"({self.node.network.node_alias(pk)}):"
            )
            # loop over channels with peer
            peer_capacity = 0
            peer_local_balance = 0
            peer_number_forwardings_out = 0
            peer_total_forwarding_out = 0
            cumul_base_fee = 0
            cumul_fee_rate = 0

            # accumulate information on a per peer basis
            for channel_id in cs:
                # collect channel info and stats
                channel_data = self.channels[channel_id]
                channel_stats = self.channels_forwarding_stats.get(channel_id, None)

                if channel_stats is None:
                    number_forwardings_out = 0
                    total_forwarding_out = 0
                else:
                    number_forwardings_out = channel_stats["number_forwardings_out"]
                    total_forwarding_out = channel_stats["total_forwarding_out"]

                peer_capacity += channel_data["capacity"]
                peer_local_balance += channel_data["local_balance"]
                peer_number_forwardings_out += number_forwardings_out
                peer_total_forwarding_out += total_forwarding_out

                cumul_fee_rate += self.channel_fee_policies[
                    channel_data["channel_point"]
                ]["fee_rate"]
                cumul_base_fee += self.channel_fee_policies[
                    channel_data["channel_point"]
                ]["base_fee_msat"]

            logger.info(
                f"    Channels with peer: {len(cs)}, "
                f"total capacity: {peer_capacity}, "
                f"total local balance: {peer_local_balance}"
            )

            # calculate average base fee and fee rate
            base_fee_msat = int(cumul_base_fee / len(cs))
            fee_rate = round(cumul_fee_rate / len(cs), 6)

            # FEE RATES
            factor_demand = delta_demand(
                self.params,
                self.time_interval_days,
                peer_total_forwarding_out,
                peer_local_balance,
                peer_capacity,
            )

            # round down to 6 digits, as this is the expected data for
            # the api
            fee_rate_new = round(fee_rate * factor_demand, 6)

            # if the fee rate is too low, cap it, as we don't want to
            # necessarily have too low fees, limit also from top
            fee_rate_new = max(self.params["min_fee_rate"], fee_rate_new)

            # if the fee rate is too high, cap it, as we don't want to
            # loose channels due to other parties thinking we are too greedy
            fee_rate_new = min(self.params["max_fee_rate"], fee_rate_new)

            # if we initialize the fee optimization, we want to start with
            # reasonable starting values
            if init:
                fee_rate_new = self.params["max_fee_rate"] / 2

            # BASE FEES
            factor_base_fee = self.factor_demand_base_fee(peer_number_forwardings_out)
            base_fee_msat_new = base_fee_msat * factor_base_fee
            if init:
                base_fee_msat_new = self.params["min_base_fee"]
            else:
                # limit from below
                base_fee_msat_new = int(
                    max(self.params["min_base_fee"], base_fee_msat_new)
                )
                # limit from above
                base_fee_msat_new = int(
                    min(self.params["max_base_fee"], base_fee_msat_new)
                )

            logger.info(
                "    Fee rate change: %1.6f -> %1.6f (factor %1.3f)",
                fee_rate,
                fee_rate_new,
                factor_demand,
            )

            logger.info(
                "    Base fee change: %4d -> %4d (factor %1.3f)",
                base_fee_msat,
                base_fee_msat_new,
                factor_base_fee,
            )

            # second loop through channels
            for channel_id in cs:
                channel_data = self.channels[channel_id]
                channel_stats = self.channels_forwarding_stats.get(channel_id, None)
                if channel_stats is None:
                    flow = 0
                    fees_sat = 0
                    total_forwarding_in = 0
                    total_forwarding_out = 0
                    number_forwardings = 0
                    number_forwardings_out = 0
                else:
                    flow = channel_stats["flow_direction"]
                    fees_sat = channel_stats["fees_out"] / 1000
                    total_forwarding_in = channel_stats["total_forwarding_in"]
                    total_forwarding_out = channel_stats["total_forwarding_out"]
                    number_forwardings = channel_stats["number_forwardings"]
                    number_forwardings_out = channel_stats["number_forwardings_out"]

                lb = channel_data["local_balance"]
                ub = channel_data["unbalancedness"]
                capacity = channel_data["capacity"]

                logger.info("  > Statistics for channel %s:", channel_id)
                logger.info(
                    "    ub: %0.2f, flow: %0.2f, fees: %1.3f sat, "
                    "cap: %d sat, lb: %d sat, nfwd: %d, in: %d sat, "
                    "out: %d sat.",
                    ub,
                    flow,
                    fees_sat,
                    capacity,
                    lb,
                    number_forwardings,
                    total_forwarding_in,
                    total_forwarding_out,
                )

                stats.append(
                    {
                        "date": self.time_end,
                        "channelid": channel_id,
                        "total_in": total_forwarding_in,
                        "total_out": total_forwarding_out,
                        "lb": lb,
                        "ub": ub,
                        "flow": flow,
                        "fees": fees_sat,
                        "cap": capacity,
                        "fdem": factor_demand,
                        "fr": self.channel_fee_policies[channel_data["channel_point"]][
                            "fee_rate"
                        ],
                        "frn": fee_rate_new,
                        "nfwd": number_forwardings,
                        "nfwdo": number_forwardings_out,
                        "fbase": factor_base_fee,
                        "bf": self.channel_fee_policies[channel_data["channel_point"]][
                            "base_fee_msat"
                        ],
                        "bfn": base_fee_msat_new,
                    }
                )

                if ignore_peer:
                    logger.info(f"    Ignore channel {channel_id} due to config file.")
                else:
                    channel_fee_policies[channel_data["channel_point"]] = {
                        "base_fee_msat": base_fee_msat_new,
                        "fee_rate": fee_rate_new,
                        "cltv": self.params["cltv"],
                    }
            logger.info("")
        return channel_fee_policies, stats

    def factor_demand_base_fee(self, num_fwd_out: int) -> float:
        """Calculates a change factor by taking into account the number of
        transactions transacted in a time interval compared to a fixed number
        of transactions.

        :param num_fwd_out: number of outward forwardings
        :return: [1-c_max, 1+c_max]"""
        logger.info("    Number of outward forwardings: %6.0f", num_fwd_out)

        n = num_fwd_out / self.time_interval_days
        delta = 1 + optimization_parameters["delta_b"] * (
            n / optimization_parameters["n_t"] - 1
        )

        delta = max(
            1 - optimization_parameters["delta_b_min"],
            min(delta, 1 + optimization_parameters["delta_b_max"]),
        )
        return delta

    def append_to_history(self, stats: List[dict]):
        """append_history adds the fee setting statistics to a pickle file.

        :param stats: fee statistics"""
        logger.debug("Saving fee setting stats to fee history.")
        with open(self.history_path, "a") as f:
            json.dump(stats, f)
            f.write("\n")

    def read_history(self) -> List[dict]:
        """read_history is a function for unpickling the fee setting history.

        :return: list of fee setting statistics
        :rtype: list[dict]"""
        with open(self.history_path, "r") as f:
            history = []
            for i, line in enumerate(f):
                history.append(json.loads(line))
        return history


if __name__ == "__main__":
    from lndmanage.lib.node import LndNode
    import logging.config

    logging.config.dictConfig(settings.logger_config)

    nd = LndNode("/home/user/.lndmanage/config.ini")
    fee_setter = FeeSetter(nd)
    fee_setter.set_fees()



from collections import defaultdict
from statistics import median, mean

import numpy as np
import networkx as nx

from lndmanage.lib.ln_utilities import convert_channel_id_to_short_channel_id
from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class NetworkAnalysis(object):
    """
    Class for network analysis.

    """

    def __init__(self, node):
        """
        :param node: :class:`lib.node.LndNode`
        """

        self.node = node
        self.nodes_info = None

    def find_nodes_with_largest_degrees(self, node_count=10):
        """
        Finds node_count nodes in the graph, which have the most connections.

        :param node_count: int
        :return: list of nodes sorted by degree
        """

        nodes_and_degrees = list(self.node.network.graph.degree)

        # number of channels in networkx is twice the real number of channels
        nodes_and_degrees = [(n[0], n[1] // 2) for n in nodes_and_degrees]

        nodes_sorted_by_degrees_decremental = sorted(
            nodes_and_degrees, key=lambda x: x[1], reverse=True)

        return nodes_sorted_by_degrees_decremental[:node_count]

    def find_nodes_with_highest_total_capacities(self, node_count=10):
        """
        Finds node_count nodes in the graph with the largest amount of bitcoin
        assigned in their channels.

        :param node_count: int
        :return: list of nodes sorted by capacity
        """

        nodes_and_capacity = []

        for n in self.node.network.graph.nodes:
            total_capacity = 0
            edges = self.node.network.graph.edges(n, data=True)
            for e in edges:
                total_capacity += e[2]['capacity']
            nodes_and_capacity.append((n, total_capacity))

        nodes_and_capacity = sorted(
            nodes_and_capacity, key=lambda x: x[1], reverse=True)

        return nodes_and_capacity[:node_count]

    def get_sorted_nodes_by_property(self, key='capacity', node_count=10,
                                     decrementing=True, min_degree=0):
        """
        Returns sorted list of nodes by the key field.

        A minimal number of degree of the target nodes can be given.

        :param key: property by which it is sorted
        :param node_count:
        :param decrementing:
        :param min_degree:
        :return: sorted list
        """

        nodes = []
        for node_info in self.nodes_info:
            if node_info['degree'] >= min_degree:
                nodes.append(node_info)

        sorted_nodes = sorted(
            nodes, key=lambda x: x[key], reverse=decrementing)

        return sorted_nodes[:node_count]

    def node_info_basic(self, node_pub_key):
        node_info = self.node.get_node_info(node_pub_key)
        # calculate average and mean channel fees

        base_fees = []
        fee_rates_milli_msat = []
        capacities = []
        for c in node_info['channels']:
            # Determine which policy to look at.
            if node_pub_key == c.node1_pub:
                policy = c.node1_policy
            else:
                policy = c.node2_policy
            base_fees.append(policy.fee_base_msat)
            fee_rates_milli_msat.append(policy.fee_rate_milli_msat)
            capacities.append(c.capacity)

        node_info['mean_base_fee'] = int(mean(base_fees))
        node_info['median_base_fee'] = int(median(base_fees))
        node_info['mean_fee_rate'] = round(mean(fee_rates_milli_msat) / 1E6, 6)
        node_info['median_fee_rate'] = round(median(fee_rates_milli_msat) / 1E6, 6)
        node_info['mean_capacity'] = int(mean(capacities))
        node_info['median_capacity'] = int(median(capacities))

        return node_info

    def node_information(self, node_pub_key):
        """
        Extracts information on a node from the networkx graph.

        :param node_pub_key: string, public key of the analyzed node
        :return: dict of properties
        """

        total_capacity = 0
        edges = self.node.network.graph.edges(node_pub_key, data=True)
        degree = len(edges)
        for e in edges:
            total_capacity += e[2]['capacity']

        return {'node_id': node_pub_key,
                'capacity': total_capacity,
                'degree': degree,
                'capacity_per_channel': total_capacity / max(1, degree),
                'user_nodes': self.number_of_connected_user_nodes(
                    node_pub_key),
                }

    def nodes_information(self):
        """
        Extract all nodes' properties from the network.

        :return: list of dicts of nodes
        """
        nodes = []
        for n in self.node.network.graph.nodes:
            node_info = self.node_information(n)
            nodes.append(node_info)
        return nodes

    def print_node_overview(self, node_pub_key):
        """
        Prints an overview of any node on the network.

        Lists the channels and their capacities/fees.

        :param node_pub_key:
        """

        logger.info("-------- Node overview for node {} --------".format(
            node_pub_key))
        edges = list(self.node.network.graph.edges(node_pub_key, data=True))
        sorted(edges, key=lambda x: x[1])
        logger.info("Node has {} channels".format(len(edges)))
        for ie, e in enumerate(edges):
            logger.info("Channel number: {} between {} and {}".format(
                ie, e[0], e[1]))
            logger.info("Channel information: {}".format(e[2]))

    def number_of_connected_user_nodes(self, node_pub_key):
        """
        Determines the number of 'user' nodes that a node is connected to.

        A user node is determined by having a smaller amount of degrees than
        a certain value NUMBER_CHANNELS_DEFINING_USER_NODE. A node with a
        low number of connections is assumed to be a user node.

        :param node_pub_key: public_key of a node to be analyzed
        :return:
        """

        connected_end_nodes = 0
        edges = self.node.network.graph.edges(node_pub_key)

        for e in edges:
            degree_neighbor = self.node.network.graph.degree(e[1]) // 2
            if degree_neighbor <= settings.NUMBER_CHANNELS_DEFINING_USER_NODE:
                connected_end_nodes += 1

        return connected_end_nodes

    def get_nodes_n_hops_away(self, node_pub_key, n):
        """
        Returns all nodes, which are n hops away from a given
        node_pub_key node.

        :param node_pub_key: string
        :param n: int
        :return: dict with nodes and distance as value
        """

        return nx.single_source_shortest_path_length(
            self.node.network.graph, node_pub_key, cutoff=n)

    def secondary_hops_added(self, node_pub_key):
        """
        Determines the number of secondary hops added if connected to the node.

        :param node_pub_key: str
        :return: int
        """
        potential_new_second_neighbors = set(
            nx.all_neighbors(self.node.network.graph, node_pub_key))
        current_close_neighbors = set(
            self.get_nodes_n_hops_away(self.node.pub_key, 2).keys())
        new_second_neighbors = potential_new_second_neighbors.difference(
            current_close_neighbors)
        return len(new_second_neighbors)

    def nodes_most_second_neighbors(self, node_pub_key, number_of_nodes=10):
        """
        Which node should be added in order to reach the most other nodes
        with two hops?

        :param node_pub_key: string
        :param number_of_nodes: int
        :return: list of results nodes, adding the most secondary neighbors
        """

        node_candidates = []

        # set of nodes currently two hops away
        current_close_neighbors = set(
            self.get_nodes_n_hops_away(node_pub_key, 2).keys())

        # loop through nodes in the network and check their direct neighbors
        for n in self.node.network.graph:
            potential_new_second_neighbors = set(
                nx.all_neighbors(self.node.network.graph, n))

            # subtract current_close_neighbors from
            # the potential new second neighbors
            new_second_neighbors = potential_new_second_neighbors.difference(
                current_close_neighbors)

            # add the node and its number of secondary neighbors to a list
            node_candidates.append([n, len(new_second_neighbors)])

        nodes_sorted = sorted(
            node_candidates, key=lambda x: x[1], reverse=True)

        return nodes_sorted[:number_of_nodes]

    def print_find_nodes_giving_most_secondary_hops(self, node_pub_key):
        """
        Determines and prints the nodes giving the most second
        nearest neighbors.

        :param node_pub_key: node public key of the interested node
        """

        nodes = self.nodes_most_second_neighbors(node_pub_key)
        logger.info("Finding all nodes, which when connected would give n new "
                    "nodes reachable with two hops.")

        for node, number_neighbors in nodes:
            logger.info(f"Node: {node} - new neighbors: {number_neighbors}")

    def determine_channel_openings(self, from_days_ago):
        """
        Determines all channel openings in the last `from_days_ago` days and
        creates a dictionary of nodes involved.

        The dictionary values contain tuples of channel creation height and
        capacity of the channels that were opened.

        :param from_days_ago: int
        :return: dict, keys: node public keys, values: (block height, capacity)
        """

        logger.info(f"Determining channel openings in the last "
                    f"{from_days_ago} days (excluding already closed ones).")
        # retrieve all channels in the network
        all_channels_list = self.node.network.edges.keys()

        # make sure the channels are sorted by age, oldest first
        all_channels_list = sorted(all_channels_list)

        # determine blockheight from where to start the analysis
        # we have about six blocks per hour
        blockheight_start = self.node.blockheight - from_days_ago * 24 * 6

        # take only youngest channels
        channels_filtered_and_creation_time = []
        for cid in all_channels_list:
            height = convert_channel_id_to_short_channel_id(cid)[0]
            if height > blockheight_start:
                channels_filtered_and_creation_time.append((cid, height))
        logger.info(f"In the last {from_days_ago} days, there were at least "
                    f"{len(channels_filtered_and_creation_time)} "
                    f"channel openings.")

        # analyze the openings and assign tuples of
        # (creation height, channel capacity) to nodes
        channel_openings_per_node_dict = defaultdict(list)
        for c, height in channels_filtered_and_creation_time:
            edge = self.node.network.edges[c]
            channel_openings_per_node_dict[edge['node1_pub']].append(
                (height, edge['capacity']))
            channel_openings_per_node_dict[edge['node2_pub']].append(
                (height, edge['capacity']))

        return channel_openings_per_node_dict

    def calculate_channel_opening_statistics(self, from_days_ago,
                                             exclude_openings_less_than=5):
        """
        Calculates basic channel opening statistics for each node.

        :param from_days_ago: int
        :param exclude_openings_less_than: int, nodes with smaller channel
            openings than this are excluded
        :return: dict, keys: nodes, values: serveral heuristics
        """

        openings_per_node_dict = self.determine_channel_openings(from_days_ago)
        opening_statistics_per_node = {}

        for n, nv in openings_per_node_dict.items():
            # convert opening characteristics (heights and capacities) to lists
            heights = [opening[0] for opening in nv]
            capacities = [opening[1] for opening in nv]

            # calculate the blockheight differences between successive
            # channel openings (tells about frequency)
            delta_heights = np.diff(heights)

            # calculate median and average differences of channel openings
            # (median can give hints on bursts of openings)
            median_opening_time = np.median(delta_heights)
            average_opening_time = np.mean(delta_heights)

            # other interesting quantities
            openings = len(capacities)
            openings_total_capacity = sum(capacities)
            openings_average_capacity = \
                float(openings_total_capacity) / openings
            node_total_capacity = self.node.network.node_capacity(n)
            node_number_channels = self.node.network.number_channels(n)

            # put all the data in a dictionary, which can then be handled
            # by the node recommendation class
            if openings > exclude_openings_less_than:
                opening_statistics_per_node[n] = {
                    'opening_median_time': median_opening_time,
                    'opening_average_time': average_opening_time,
                    'openings_average_capacity':  # unit: btc
                        openings_average_capacity / 1E8,
                    'openings': openings,
                    'openings_total_capacity':  # unit: btc
                        openings_total_capacity / 1E8,
                    'relative_openings':
                        float(openings) / node_number_channels,
                    'relative_total_capacity':  # unit: ksat
                        0 if node_total_capacity == 0 else float(openings_total_capacity) / node_total_capacity,
                }

        return opening_statistics_per_node

    def distance(self, first_node, second_node):
        """
        Calculates the distance in hops from first node to second node.
        :param first_node: str
        :param second_node: str
        :return: int
        """
        try:
            distance = nx.shortest_path_length(
                self.node.network.graph, source=first_node, target=second_node)
        except nx.exception.NetworkXNoPath:
            distance = float('inf')  # some high number

        return distance


if __name__ == '__main__':
    from lndmanage.lib.node import LndNode

    import logging.config
    logging.config.dictConfig(settings.logger_config)

    nd = LndNode()
    network_analysis = NetworkAnalysis(nd)

    nodes_capacities = network_analysis.find_nodes_with_highest_total_capacities()
    network_analysis.print_find_nodes_giving_most_secondary_hops(nd.pub_key)



"""Module for gathering statistics of channels or nodes."""
from collections import OrderedDict, defaultdict
import logging
from typing import Dict

import numpy as np

from lndmanage.lib.data_types import NodeProperties
from lndmanage.lib.node import LndNode
from lndmanage.lib.ln_utilities import local_balance_to_unbalancedness
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

np.warnings.filterwarnings("ignore")

# nearest neighbor weight for flow-analysis ~1/(avg. degree)
NEIGHBOR_WEIGHT = 0.1
# nearest neighbor weight for flow-analysis ~1/(avg. degree)^2
NEXT_NEIGHBOR_WEIGHT = 0.02


def nan_to_zero(number: float) -> float:
    if number is np.nan or number != number:
        return 0.0
    else:
        return number


class ForwardingAnalyzer(object):
    """Analyzes forwardings for single channels."""

    def __init__(self, node: "LndNode"):
        self.node = node
        self.forwarding_events = self.node.get_forwarding_events()
        self.channel_forwarding_stats = {}  # type: Dict[str, ForwardingStatistics]
        self.node_forwarding_stats = {}  # type: Dict[str, ForwardingStatistics]
        self.max_time_interval_days = None

    def initialize_forwarding_stats(self, time_start: float, time_end: float):
        """Initializes the channel and node statistics objects with data from forwardings.

        :param time_start: time interval start, unix timestamp
        :param time_end: time interval end, unix timestamp
        """
        channel_id_to_node_id = self.node.channel_id_to_node_id()
        self.channel_forwarding_stats = defaultdict(ForwardingStatistics)
        self.node_forwarding_stats = defaultdict(ForwardingStatistics)

        min_timestamp = float("inf")
        max_timestamp = 0
        for f in self.forwarding_events:
            if time_start < f["timestamp"] < time_end:
                # find min and max range of forardings
                if f["timestamp"] > max_timestamp:
                    max_timestamp = f["timestamp"]
                if f["timestamp"] < min_timestamp:
                    min_timestamp = f["timestamp"]
                # make a dictionary entry for unknown channels
                channel_id_in = f["chan_id_in"]
                channel_id_out = f["chan_id_out"]
                node_id_in = channel_id_to_node_id.get(channel_id_in)
                node_id_out = channel_id_to_node_id.get(channel_id_out)

                # node statistics
                if node_id_in:
                    self.node_forwarding_stats[node_id_in].inward_forwardings.append(
                        f["amt_in"]
                    )
                    self.node_forwarding_stats[node_id_in].fees_in.append(
                        f["fee_msat"]
                    )
                if node_id_out:
                    self.node_forwarding_stats[node_id_out].outward_forwardings.append(
                        f["amt_out"]
                    )
                    self.node_forwarding_stats[node_id_out].fees_out.append(
                        f["fee_msat"]
                    )
                    self.node_forwarding_stats[node_id_out].timestamps.append(
                        f["timestamp"]
                    )

                # channel statistics
                # inward channel:
                self.channel_forwarding_stats[channel_id_in].inward_forwardings.append(
                    f["amt_in"]
                )
                self.channel_forwarding_stats[channel_id_in].fees_in.append(
                    f["fee_msat"]
                )
                # outward channel:
                self.channel_forwarding_stats[
                    channel_id_out
                ].outward_forwardings.append(f["amt_out"])
                self.channel_forwarding_stats[channel_id_out].fees_out.append(
                    f["fee_msat"]
                )
                self.channel_forwarding_stats[channel_id_out].timestamps.append(
                    f["timestamp"]
                )

        # determine the time interval starting with the first forwarding
        # to the last forwarding in the analyzed time interval determined
        # by time_start and time_end
        self.max_time_interval_days = (max_timestamp - min_timestamp) / (24 * 60 * 60)

    def get_forwarding_statistics_channels(self):
        """Prepares the forwarding statistics for each channel.

        :return: dict: statistics with channel_id as keys
        """
        channel_statistics = {}

        for k, c in self.channel_forwarding_stats.items():
            channel_statistics[k] = {
                "fees_out": c.total_fees_out(),
                "fees_in": c.total_fees_in(),
                "fees_in_out": c.total_fees_in() + c.total_fees_out(),
                "flow_direction": c.flow_direction(),
                "mean_forwarding_in": c.mean_forwarding_in(),
                "mean_forwarding_out": c.mean_forwarding_out(),
                "median_forwarding_in": c.median_forwarding_in(),
                "median_forwarding_out": c.median_forwarding_out(),
                "number_forwardings": c.number_forwardings(),
                "number_forwardings_out": c.number_forwardings_out(),
                "largest_forwarding_amount_in": c.largest_forwarding_amount_in(),
                "largest_forwarding_amount_out": c.largest_forwarding_amount_out(),
                "total_forwarding_in": c.total_forwarding_in(),
                "total_forwarding_out": c.total_forwarding_out(),
            }
        return channel_statistics

    def get_forwarding_statistics_nodes(self, sort_by="total_forwarding"):
        """Calculates forwarding statistics based on individual channels for
        their nodes.

        :param sort_by: str, abbreviation for the dict key
        :return:
        """
        closed_channels = self.node.get_closed_channels()
        open_channels = self.node.get_open_channels()
        logger.debug(
            f"Number of channels with known forwardings: "
            f"{len(closed_channels) + len(open_channels)} "
            f"(thereof {len(closed_channels)} closed channels)."
        )

        forwarding_stats = defaultdict(dict)
        for nid, node_stats in self.node_forwarding_stats.items():
            tot_in = node_stats.total_forwarding_in()
            tot_out = node_stats.total_forwarding_out()
            forwarding_stats[nid]["fees_out"] = node_stats.total_fees_out()
            forwarding_stats[nid]["fees_in"] = node_stats.total_fees_in()
            forwarding_stats[nid]["fees_in_out"] = node_stats.total_fees_in() + node_stats.total_fees_out()
            forwarding_stats[nid]["flow_direction"] = (
                -((float(tot_in) / (tot_in + tot_out)) - 0.5) / 0.5
            )
            forwarding_stats[nid][
                "largest_forwarding_amount_in"
            ] = node_stats.largest_forwarding_amount_in()
            forwarding_stats[nid][
                "largest_forwarding_amount_out"
            ] = node_stats.largest_forwarding_amount_out()
            forwarding_stats[nid][
                "mean_forwarding_in"
            ] = node_stats.mean_forwarding_in()
            forwarding_stats[nid][
                "mean_forwarding_out"
            ] = node_stats.mean_forwarding_out()
            forwarding_stats[nid][
                "median_forwarding_in"
            ] = node_stats.median_forwarding_in()
            forwarding_stats[nid][
                "median_forwarding_out"
            ] = node_stats.median_forwarding_out()
            forwarding_stats[nid][
                "number_forwardings"
            ] = node_stats.number_forwardings()
            forwarding_stats[nid]["total_forwarding_in"] = tot_in
            forwarding_stats[nid]["total_forwarding_out"] = tot_out
            forwarding_stats[nid]["total_forwarding"] = tot_in + tot_out

        sorted_dict = OrderedDict(
            sorted(forwarding_stats.items(), key=lambda x: -x[1][sort_by])
        )

        return sorted_dict

    def simple_flow_analysis(self, last_forwardings_to_analyze=100):
        """Takes each forwarding event and determines the set of incoming nodes
        (up to second nearest neighbors) and outgoing nodes
        (up to second nearest neighbors) and does a frequency analysis of both
        sets, assigning a probability that a certain node was involved in
        the forwarding process.

        These probabilities are then used to take the difference between
        both, excluding effectively the large hubs. The sets can be weighted
        by a quantity of the forwarding process, e.g. the forwarding fee.

        Two lists are returned, the rest of the incoming and the rest of the
        outgoing node sets.

        :param last_forwardings_to_analyze:
            number of last forwardings to be analyzed
        :type last_forwardings_to_analyze: int

        :return: sending list, receiving list"""
        # TODO: refine with channel capacities
        # TODO: refine with update time
        # TODO: refine with route calculation
        # TODO: fine tune NEIGHBOR_WEIGHT, NEXT_NEIGHBOR_WEIGHT

        logger.info("Doing simple forwarding analysis.")

        # initialize node score dictionaries
        total_incoming_neighbors = defaultdict(float)
        total_outgoing_neighbors = defaultdict(float)

        logger.info(
            f"Total forwarding events found: " f"{len(self.forwarding_events)}."
        )
        logger.info(
            f"Carrying out flow analysis for last "
            f"{last_forwardings_to_analyze} forwarding events."
        )
        number_progress_report = last_forwardings_to_analyze // 10

        meaningful_outward_forwardings = 0
        for nf, f in enumerate(self.forwarding_events[-last_forwardings_to_analyze:]):

            # report progress
            if nf % number_progress_report == 0:
                logger.info(
                    f"Analysis progress: "
                    f"{100 * float(nf) / last_forwardings_to_analyze}%"
                )

            chan_id_in = f["chan_id_in"]
            chan_id_out = f["chan_id_out"]

            edge_data_in = self.node.network.edges.get(chan_id_in, None)
            edge_data_out = self.node.network.edges.get(chan_id_out, None)

            if edge_data_in is not None and edge_data_out is not None:
                # determine incoming and outgoing node pub keys
                incoming_node_pub_key = (
                    edge_data_in["node1_pub"]
                    if edge_data_in["node1_pub"] != self.node.pub_key
                    else edge_data_in["node2_pub"]
                )
                outgoing_node_pub_key = (
                    edge_data_out["node1_pub"]
                    if edge_data_out["node1_pub"] != self.node.pub_key
                    else edge_data_out["node2_pub"]
                )

                # nodes involved in the forwarding process should be excluded
                excluded_nodes = [
                    self.node.pub_key,
                    incoming_node_pub_key,
                    outgoing_node_pub_key,
                ]

                # determine all the nearest and second nearest
                # neighbors of the incoming/outgoing nodes,
                # they may appear more than once
                incoming_neighbors = self.__determine_joined_neighbors(
                    incoming_node_pub_key, excluded_nodes=excluded_nodes
                )
                outgoing_neighbors = self.__determine_joined_neighbors(
                    outgoing_node_pub_key, excluded_nodes=excluded_nodes
                )

                # do a symmetric difference of node sets with weights
                symmetric_difference_weights = self.__symmetric_difference(
                    incoming_neighbors, outgoing_neighbors
                )

                final_outgoing_nodes = self.__filter_nodes(
                    symmetric_difference_weights, return_positive_weights=True
                )
                final_incoming_nodes = self.__filter_nodes(
                    symmetric_difference_weights, return_positive_weights=False
                )

                # normalize the weights
                normalized_incoming = self.__normalize_neighbors(final_incoming_nodes)
                normalized_outgoing = self.__normalize_neighbors(final_outgoing_nodes)

                # set weight for each forwarding event
                weight = 1
                # alternatively:
                # weight = f['amt_in']
                # weight = f['fee_msat']

                for n, nv in normalized_incoming.items():
                    total_incoming_neighbors[n] += nv * weight

                # If we know, that the outward hop already reached the
                # target of the payment, we don't want to add the neighbors
                # of the outgoing node to the total statistics.
                # We know if the next hop was the final one of the payment by
                # testing whether the forwarding amount in msat has remainder
                # zero when divided by 1000 or not, provided the sent amount
                # was larger than 1E6 msat.
                if f["amt_out_msat"] % 1000:
                    meaningful_outward_forwardings += 1
                    logger.debug(
                        f"Forwarding was not last hop: {f['amt_out_msat']}, "
                        f"chan_id_out: {chan_id_out}"
                    )
                    for n, nv in normalized_outgoing.items():
                        total_outgoing_neighbors[n] += nv * weight
        logger.info(
            f"Could use {meaningful_outward_forwardings} "
            f"forwardings to estimate targets of payments."
        )

        # sort according to weights
        total_incoming_node_dict = self.__weighted_neighbors_to_sorted_dict(
            total_incoming_neighbors
        )
        total_outgoing_node_dict = self.__weighted_neighbors_to_sorted_dict(
            total_outgoing_neighbors
        )

        return total_incoming_node_dict, total_outgoing_node_dict

    @staticmethod
    def __weighted_neighbors_to_sorted_dict(node_dict):
        """
        Converts a node weight dictionary to a sorted list of
        nodes with weights.

        :param node_dict: dict with key-value pairs of node_pub_key and weight
        :return: sorted list
        """
        sorted_nodes_dict = OrderedDict()
        node_list = [(n, nv) for n, nv in node_dict.items()]
        node_list_sorted = sorted(node_list, key=lambda x: x[1], reverse=True)
        for n, nv in node_list_sorted:
            sorted_nodes_dict[n] = {"weight": nv}
        return sorted_nodes_dict

    def __determine_joined_neighbors(self, node_pub_key, excluded_nodes):
        """Determines the joined set of nearest and second neighbors and assigns
        a weight to every node dependent how often they appear.

        :param node_pub_key: str, public key of the home node
        :param excluded_nodes: list of str, public keys of excluded nodes
                                            in analysis
        :return: dict, keys: node_pub_keys, values: weights
        """

        neighbors = list(self.node.network.neighbors(node_pub_key))
        second_neighbors = list(self.node.network.second_neighbors(node_pub_key))

        # determine neighbor node_weights
        neighbor_weights = self.__analyze_neighbors(
            neighbors, excluded_nodes=excluded_nodes, weight=NEIGHBOR_WEIGHT
        )
        second_neighbor_weights = self.__analyze_neighbors(
            second_neighbors, excluded_nodes=excluded_nodes, weight=NEXT_NEIGHBOR_WEIGHT
        )

        # combine nearest and second nearest neighbor node weights
        joined_neighbors = self.__join_neighbors(
            neighbor_weights, second_neighbor_weights
        )

        return joined_neighbors

    @staticmethod
    def __normalize_neighbors(neighbors_weight_dict):
        """
        Normalizes the weights of the neighbors to the total weight
        of all neighbors.

        :param neighbors_weight_dict: dict, keys: node_pub_keys,
                                            values: weights
        :return: dict, keys: node_pub_keys, values: normalized weights
        """
        total_weight = 0
        normalized_dict = {}
        # determine total weight
        for n, nv in neighbors_weight_dict.items():
            total_weight += nv
        # normalize
        for n, nv in neighbors_weight_dict.items():
            normalized_dict[n] = float(nv) / total_weight

        return normalized_dict

    @staticmethod
    def __analyze_neighbors(neighbors, excluded_nodes, weight):
        """
        Analyzes a node dict for the frequency of nodes and gives them
        a weight. An upper bound of the weights is set.

        :param neighbors: list of node_pub_keys
        :param excluded_nodes: excluded node_pub_keys for analysis
        :param weight: float, weight for each individual appearance of a node
        :return: dict, keys: node_pub_keys, values: node weights
        """
        node_weights = {}

        for n in neighbors:
            if n not in excluded_nodes:
                if n in node_weights.keys():
                    node_weights[n] = min(node_weights[n] + weight, 1.0)
                else:
                    node_weights[n] = weight

        return node_weights

    @staticmethod
    def __join_neighbors(first_neighbor_dict, second_neighbor_dict):
        """Joins two node weight dicts together.
        :param first_neighbor_dict: dict, keys: node_pub_keys,
                                          values: node weights
        :param second_neighbor_dict: dict, keys: node_pub_keys,
                                           values: node weights
        :return: dict, keys: node_pub_keys, values: node weights
        """
        # make a copy of the first node weight dict
        joined_neighbor_dict = dict(first_neighbor_dict)

        # add all the nodes from the second dict
        for n, v in second_neighbor_dict.items():
            if n in joined_neighbor_dict:
                joined_neighbor_dict[n] = min(
                    1, joined_neighbor_dict[n] + second_neighbor_dict[n]
                )
            else:
                joined_neighbor_dict[n] = second_neighbor_dict[n]

        return joined_neighbor_dict

    @staticmethod
    def __symmetric_difference(first_neighbors_dict, second_neighbors_dict):
        """
        Calculates the difference of weights of first and second node dicts,
        doing also a symmetric difference between the sets of nodes.

        :param first_neighbors_dict: dict, keys: node_pub_keys,
                                           values: node weights
        :param second_neighbors_dict: dict, keys: node_pub_keys,
                                            values: node weights
        :return: dict, keys: node_pub_keys, values: node weights
        """
        first_nodes = set(first_neighbors_dict.keys())
        second_nodes = set(second_neighbors_dict.keys())

        nodes_intersection = first_nodes.intersection(second_nodes)

        unique_first_nodes = first_nodes - nodes_intersection
        unique_second_nodes = second_nodes - nodes_intersection

        diff_neighbor_dict = {}
        for n in unique_first_nodes:
            diff_neighbor_dict[n] = first_neighbors_dict[n]

        for n in unique_second_nodes:
            diff_neighbor_dict[n] = second_neighbors_dict[n]

        for n in nodes_intersection:
            delta = first_neighbors_dict[n] - second_neighbors_dict[n]
            if delta != 0.0:
                diff_neighbor_dict[n] = delta

        return diff_neighbor_dict

    @staticmethod
    def __filter_nodes(node_weights, return_positive_weights=True):
        """
        Filters out nodes with positive or negative weights and
        takes the absolute.

        :param node_weights: dict
        :param return_positive_weights: bool, if True returns nodes with
                                              positive weights,
                                              if False negative weights
        :return:
        """
        new_node_weights = {}
        for n, nv in node_weights.items():
            if nv > 0 and return_positive_weights:
                new_node_weights[n] = nv
            elif nv < 0 and not return_positive_weights:
                new_node_weights[n] = -nv
        return new_node_weights


class ForwardingStatistics(object):
    """Functionality to analyze the forwardings of a single node/channel."""

    def __init__(self):
        self.inward_forwardings = []
        self.outward_forwardings = []
        self.timestamps = []
        self.fees_out = []
        self.fees_in = []

    def total_forwarding_in(self) -> int:
        return sum(self.inward_forwardings)

    def total_forwarding_out(self) -> int:
        return sum(self.outward_forwardings)

    def mean_forwarding_in(self) -> float:
        return np.mean(self.inward_forwardings)

    def mean_forwarding_out(self) -> float:
        return np.mean(self.outward_forwardings)

    def median_forwarding_in(self) -> float:
        return np.median(self.inward_forwardings)

    def median_forwarding_out(self) -> float:
        return np.median(self.outward_forwardings)

    def total_fees_out(self) -> int:
        return sum(self.fees_out)

    def total_fees_in(self) -> int:
        return sum(self.fees_in)

    def largest_forwarding_amount_out(self) -> int:
        return max(self.outward_forwardings, default=float("nan"))

    def largest_forwarding_amount_in(self) -> int:
        return max(self.inward_forwardings, default=float("nan"))

    def flow_direction(self) -> float:
        total_in = self.total_forwarding_in()
        total_out = self.total_forwarding_out()
        try:
            return (-total_in + total_out) / (total_in + total_out)
        except ZeroDivisionError:
            return 0

    def number_forwardings(self) -> int:
        return len(self.inward_forwardings) + len(self.outward_forwardings)

    def number_forwardings_out(self):
        return len(self.outward_forwardings)


def get_node_properites(
    node: LndNode, time_interval_start: float, time_interval_end: float
) -> Dict:
    """Joins data from channels and fwdinghistory to have extended
    information about a node.

    :return: dict of node information with channel_id as keys
    """
    forwarding_analyzer = ForwardingAnalyzer(node)
    forwarding_analyzer.initialize_forwarding_stats(
        time_interval_start, time_interval_end
    )
    node_forwarding_statistics = forwarding_analyzer.get_forwarding_statistics_nodes()
    logger.debug(
        f"Time interval (between first and last forwarding) is "
        f"{forwarding_analyzer.max_time_interval_days:6.2f} days."
    )
    channel_id_to_node_id = node.channel_id_to_node_id(open_only=True)
    node_ids_with_open_channels = {nid for nid in channel_id_to_node_id.values()}
    open_channels = node.get_open_channels()

    nodes_properties = {}  # type: Dict[str, NodeProperties]

    # for each channel, accumulate properties in node properties
    for k, c in open_channels.items():
        remote_pubkey = c["remote_pubkey"]
        try:
            properties = nodes_properties[remote_pubkey]
        except KeyError:
            nodes_properties[remote_pubkey] = NodeProperties(
                age=c["age"],
                local_fee_rates=[c["local_fee_rate"]],
                local_base_fees=[c["local_base_fee"]],
                local_balances=[c["local_balance"]],
                number_active_channels=1 if c["active"] else 0,
                number_channels=1,
                number_private_channels=1 if c["private"] else 0,
                remote_fee_rates=[c["peer_fee_rate"]],
                remote_base_fees=[c["peer_base_fee"]],
                remote_balances=[c["remote_balance"]],
                sent_received_per_week=c["sent_received_per_week"],
                public_capacities=[c["capacity"]] if not c["private"] else [],
                private_capacites=[c["capacity"]] if c["private"] else [],
            )
        else:
            properties.age = max(c["age"], nodes_properties[c["remote_pubkey"]].age)
            properties.local_fee_rates.append(c["local_fee_rate"])
            properties.local_base_fees.append(c["local_base_fee"])
            properties.local_balances.append(c["local_balance"])
            properties.number_active_channels += 1 if c["active"] else 0
            properties.number_channels += 1
            properties.number_private_channels += 1 if c["private"] else 0
            properties.remote_fee_rates.append(c["peer_fee_rate"])
            properties.remote_base_fees.append(c["peer_base_fee"])
            properties.remote_balances.append(c["remote_balance"])
            properties.sent_received_per_week += c["sent_received_per_week"]
            if not c["private"]:
                properties.public_capacities.append(c["capacity"])
            else:
                properties.private_capacites.append(c["capacity"])

    # unify node properties with forwarding data
    node_properties_forwardings = {}
    # we start with looping over node properties, as this info is complete
    for node_id, properties in nodes_properties.items():
        local_balance = sum(properties.local_balances)
        remote_balance = sum(properties.remote_balances)
        capacity = sum(properties.private_capacites) + sum(properties.public_capacities)
        # there can be old forwarding data, which we neglect
        if node_id not in node_ids_with_open_channels:
            continue

        # initial data:
        node_properties_forwardings[node_id] = {
            "age": properties.age,
            "alias": node.network.node_alias(node_id),
            "local_base_fee": np.median(properties.local_base_fees),
            "local_fee_rate": np.median(properties.local_fee_rates),
            "local_balance": local_balance,
            "max_local_balance": max(properties.local_balances),
            "max_remote_balance": max(properties.remote_balances),
            "number_channels": properties.number_channels,
            "number_active_channels": properties.number_active_channels,
            "number_private_channels": properties.number_private_channels,
            "node_id": node_id,
            "remote_base_fee": np.median(properties.remote_base_fees),
            "remote_fee_rate": np.median(properties.remote_fee_rates),
            "remote_balance": remote_balance,
            "sent_reveived_per_week": properties.sent_received_per_week,
            "total_capacity": capacity,
            "max_public_capacity": max(properties.public_capacities)
            if properties.public_capacities
            else 0,
            "unbalancedness": local_balance_to_unbalancedness(
                local_balance, capacity, 0, False
            )[0],
        }

        # add forwarding data if available:
        try:
            statistics = node_forwarding_statistics[node_id]
        except KeyError:  # we don't have forwarding data, populate with defaults
            node_properties_forwardings[node_id].update(
                {
                    "fees_out": 0,
                    "fees_in": 0,
                    "fees_in_out": 0,
                    "flow_direction": float("nan"),
                    "largest_forwarding_amount_in": float("nan"),
                    "largest_forwarding_amount_out": float("nan"),
                    "median_forwarding_out": float("nan"),
                    "median_forwarding_in": float("nan"),
                    "mean_forwarding_out": float("nan"),
                    "mean_forwarding_in": float("nan"),
                    "number_forwardings": 0,
                    "total_forwarding_in": 0,
                    "total_forwarding_out": 0,
                    "total_forwarding": 0,
                }
            )
        else:
            node_properties_forwardings[node_id].update(**statistics)

        try:
            node_properties_forwardings[node_id][
                "fees_out_per_week"
            ] = node_properties_forwardings[node_id]["fees_out"] / (
                forwarding_analyzer.max_time_interval_days / 7
            )
        except ZeroDivisionError:
            node_properties_forwardings[node_id]["fees_out_per_week"] = float("nan")
        try:
            node_properties_forwardings[node_id][
                "fees_in_per_week"
            ] = node_properties_forwardings[node_id]["fees_in"] / (
                    forwarding_analyzer.max_time_interval_days / 7
            )
        except ZeroDivisionError:
            node_properties_forwardings[node_id]["fees_in_per_week"] = float("nan")

    # TODO: unify with information from liquidity hints

    return node_properties_forwardings


def get_channel_properties(
    node: LndNode, time_interval_start: float, time_interval_end: float
) -> Dict:
    """Joins data from listchannels and fwdinghistory to have extended
    information about channels.

    :return: dict of channel information with channel_id as keys
    """
    forwarding_analyzer = ForwardingAnalyzer(node)
    forwarding_analyzer.initialize_forwarding_stats(
        time_interval_start, time_interval_end
    )

    # dict with channel_id keys
    statistics = forwarding_analyzer.get_forwarding_statistics_channels()
    logger.debug(
        f"Time interval (between first and last forwarding) is "
        f"{forwarding_analyzer.max_time_interval_days:6.2f} days."
    )
    # join the two data sets:
    channels = node.get_unbalanced_channels(unbalancedness_greater_than=0.0)

    for k, c in channels.items():
        # we may not have forwarding data for every channel
        chan_stats = statistics.get(c["chan_id"], {})
        c["forwardings_per_channel_age"] = (
            chan_stats.get("number_forwardings", 0.01) / c["age"]
        )
        c["bandwidth_demand"] = (
            max(
                nan_to_zero(chan_stats.get("mean_forwarding_in", 0)),
                nan_to_zero(chan_stats.get("mean_forwarding_out", 0)),
            )
            / c["capacity"]
        )
        c["fees_out"] = chan_stats.get("fees_out", 0)
        c["fees_in"] = chan_stats.get("fees_in", 0)
        c["fees_in_out"] = chan_stats.get("fees_in_out", 0)
        try:
            c["fees_out_per_week"] = chan_stats.get("fees_out", 0) / (
                forwarding_analyzer.max_time_interval_days / 7
            )
        except ZeroDivisionError:
            c["fees_out_per_week"] = float("nan")
        try:
            c["fees_in_per_week"] = chan_stats.get("fees_in", 0) / (
                    forwarding_analyzer.max_time_interval_days / 7
            )
        except ZeroDivisionError:
            c["fees_in_per_week"] = float("nan")
        c["flow_direction"] = chan_stats.get("flow_direction", float("nan"))
        c["median_forwarding_in"] = chan_stats.get("median_forwarding_in", float("nan"))
        c["median_forwarding_out"] = chan_stats.get(
            "median_forwarding_out", float("nan")
        )
        c["mean_forwarding_in"] = chan_stats.get("mean_forwarding_in", float("nan"))
        c["mean_forwarding_out"] = chan_stats.get("mean_forwarding_out", float("nan"))
        c["number_forwardings"] = chan_stats.get("number_forwardings", 0)
        c["largest_forwarding_amount_in"] = chan_stats.get(
            "largest_forwarding_amount_in", float("nan")
        )
        c["largest_forwarding_amount_out"] = chan_stats.get(
            "largest_forwarding_amount_out", float("nan")
        )
        c["total_forwarding_in"] = chan_stats.get("total_forwarding_in", 0)
        c["total_forwarding_out"] = chan_stats.get("total_forwarding_out", 0)

        # action required if flow same direction as unbalancedness
        # or bandwidth demand too high
        # TODO: refine 'action_required' by better metric
        if (
            c["unbalancedness"] * c["flow_direction"] > 0
            and abs(c["unbalancedness"]) > settings.UNBALANCED_CHANNEL
        ):
            c["action_required"] = True
        else:
            c["action_required"] = False
        if c["bandwidth_demand"] > 0.5:
            c["action_required"] = True

    return channels


"""Contains Lightning network specific conversion utilities."""
from typing import Tuple

import re
import time

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def convert_short_channel_id_to_channel_id(blockheight, transaction, output):
    """
    Converts short channel id (blockheight:transaction:output) to a long integer channel id.

    :param blockheight:
    :param transaction: Number of transaction in the block.
    :param output: Number of output in the transaction.
    :return: channel id: Encoded integer number representing the channel,
     can be decoded by :func:`lib.conversion.extract_short_channel_id_from_string`.
    """
    return blockheight << 40 | transaction << 16 | output


def convert_channel_id_to_short_channel_id(channel_id):
    """
    Converts a channel id to blockheight, transaction, output
    """
    return channel_id >> 40, channel_id >> 16 & 0xFFFFFF, channel_id & 0xFFFF


def extract_short_channel_id_from_string(string):
    """
    Parses a payment error message for the short channel id of the form XXXX:XXXX:XXX.

    :param string:
    :return: short channel id [blockheight, transaction, output]
    """
    match = re.search(r'[0-9]+:[0-9]+:[0-9]+', string)
    group = match.group()
    groups = list(map(int, group.split(':')))
    assert len(groups) == 3
    return groups


def local_balance_to_unbalancedness(local_balance: int, capacity: int, commit_fee: int,
                                    initiator: bool) -> Tuple[float, int]:
    """Calculates the unbalancedness.

    :return: float:
        in [-1.0, 1.0]
    """
    # inverse of the formula:
    commit_fee = 0 if not initiator else commit_fee
    return -(2 * (local_balance + commit_fee) / capacity - 1), commit_fee


def unbalancedness_to_local_balance(unbalancedness: float, capacity: int, commit_fee: int, initiator: bool) -> Tuple[int, int]:
    commit_fee = 0 if not initiator else commit_fee
    return -int(capacity * (unbalancedness - 1) / 2) - commit_fee, commit_fee


def height_to_timestamp(node, close_height):
    now = time.time()
    blocks_ago = node.blockheight - close_height
    time_ago = blocks_ago * 10 * 60
    timestamp_sec = int(now - time_ago)
    return timestamp_sec


def parse_nodeid_channelid(info: str) -> Tuple[int, str]:
    """Parse whether info contains a channel id or node public key and hand
    it back. If no info could be extracted, raise a ValueError.

    :return: channel_id, node_pub_key
    """
    exp_channel_id = re.compile("^[0-9]{13,20}$")
    exp_short_channel_id = re.compile("^[0-9]{6}x[0-9]{3}x[0-9]$")
    exp_chan_point = re.compile("^[a-z0-9]{64}:[0-9]$")
    exp_node_id = re.compile("^[a-z0-9]{66}$")

    channel_id = None
    node_pub_key = None

    # prepare input string info
    info = str(info)
    if exp_channel_id.match(info) is not None:
        logger.debug("Info represents channel id.")
        channel_id = int(info)
    elif exp_short_channel_id.match(info) is not None:
        logger.debug("Info represents short channel id.")
        # TODO: convert short channel id to channel id
        channel_id = 0
    elif exp_chan_point.match(info) is not None:
        # TODO: convert chan point to channel id
        logger.debug("Info represents short channel id.")
        channel_id = 0
    elif exp_node_id.match(info) is not None:
        logger.debug("Info represents node public key.")
        node_pub_key = info
    else:
        raise ValueError("Info string doesn't match any pattern.")

    return channel_id, node_pub_key